import logging

from keg.db import db
from keg_elements.db.mixins import DefaultColsMixin, MethodsMixin
import keg_auth
import sqlalchemy as sa

from keg_auth_ta.extensions import auth_entity_registry

log = logging.getLogger(__name__)


class EntityMixin(DefaultColsMixin, MethodsMixin):
    pass


class UserEmailMixin(keg_auth.UserEmailMixin, keg_auth.UserMixin):
    pass


@auth_entity_registry.register_user
class User(UserEmailMixin, EntityMixin, db.Model):
    __tablename__ = 'users'

    name = sa.Column(sa.Unicode)


class UserNoEmail(keg_auth.UserMixin, EntityMixin, db.Model):
    __tablename__ = 'users_no_email'

    # this model will not be assigned the relationships the form expects to be present
    permissions = []
    bundles = []
    groups = []


class UserWithToken(keg_auth.UserTokenMixin, UserEmailMixin, EntityMixin, db.Model):
    __tablename__ = 'users_with_token'


@auth_entity_registry.register_attempt
class Attempt(keg_auth.AttemptMixin, EntityMixin, db.Model):
    __tablename__ = 'attempts'


@auth_entity_registry.register_permission
class Permission(keg_auth.PermissionMixin, EntityMixin, db.Model):
    __tablename__ = 'permissions'

    def __repr__(self):
        return '<Permission id={} token={}>'.format(self.id, self.token)


@auth_entity_registry.register_bundle
class Bundle(keg_auth.BundleMixin, EntityMixin, db.Model):
    __tablename__ = 'bundles'


@auth_entity_registry.register_group
class Group(keg_auth.GroupMixin, EntityMixin, db.Model):
    __tablename__ = 'groups'



import inspect

import flask
import flask_login
from keg.web import validate_arguments, ArgumentValidationError, ViewArgumentError

from keg_auth.extensions import lazy_gettext as _
from keg_auth.model import utils as model_utils


class RequiresUser(object):
    """ Require a user to be authenticated before proceeding to decorated target. May be
        used as a class decorator or method decorator.

        Usage: @requires_user

        Note: if using along with a route decorator (e.g. Blueprint.route), requires_user
            should be the closest decorator to the method

        Examples:
        - @requires_user
        - @requires_user()
        - @requires_user(on_authentication_failure=lambda: flask.abort(400))
        - @requires_user(http_methods_excluded=['OPTIONS'])
    """
    def __init__(self, on_authentication_failure=None, http_methods_excluded=None):
        # defaults for these handlers are provided, but may be overridden here
        self._on_authentication_failure = on_authentication_failure
        self.http_methods_excluded = http_methods_excluded

    def __call__(self, class_or_function):
        # decorator may be applied to a class or a function, but the effect is different
        if inspect.isclass(class_or_function):
            if issubclass(class_or_function, flask.Blueprint):
                return self.decorate_blueprint(class_or_function)
            return self.decorate_class(class_or_function)
        return self.decorate_function(class_or_function)

    def store_auth_info(self, obj):
        obj.__keg_auth_requires_user__ = True

    def decorate_blueprint(self, bp):
        # when decorating a blueprint, we simply need to attach a before_request method
        old_init = getattr(bp, '__init__')

        def new_init(*args, **kwargs):
            old_init(*args, **kwargs)

            this = args[0]
            this.before_request(lambda: self.check_auth(instance=this))

        bp.__init__ = new_init
        self.store_auth_info(bp)
        return bp

    def decorate_class(self, cls):
        # when decorating a view class, all of the class's route methods will submit to the given
        #   auth. The view may already have check_auth defined, though, so make sure we still call
        #   it.
        method_name, old_method = next((
            (method_name, getattr(cls, method_name, None))
            for method_name in ['check_auth', 'dispatch_request']
            if callable(getattr(cls, method_name, None))
        ), (None, None))

        if not old_method:
            raise TypeError('Class must inherit from a Keg or Flask view')

        def new_method(*args, **kwargs):
            self.check_auth(instance=args[0])
            try:
                # validate_arguments is made for a function, not a class method
                # so we need to "trick" it by sending self here, but then
                # removing it before the bound method is called below
                pass_args, pass_kwargs = validate_arguments(old_method, args, kwargs.copy())
            except ArgumentValidationError as e:
                msg = _('Argument mismatch occurred: method=%s, missing=%s, '
                        'extra_keys=%s, extra_pos=%s.'
                        '  Arguments available: %s') % (old_method, e.missing,
                                                        e.extra, e.extra_positional,
                                                        kwargs)  # pragma: no cover
                raise ViewArgumentError(msg)  # pragma: no cover

            return old_method(*pass_args, **pass_kwargs)

        setattr(cls, method_name, new_method)

        # store auth info on the class itself
        self.store_auth_info(cls)
        return cls

    def decorate_function(self, func):
        # when decorating a function, we wrap it to check the auth first, then call the original
        #   function. Set the name on the wrapper for it to be available when assigning a route
        def wrapper(*args, **kwargs):
            self.check_auth()
            return func(*args, **kwargs)
        wrapper.__name__ = getattr(func, '__name__', 'wrapper')
        wrapper.__keg_auth_original_function__ = func

        # store auth info on the wrapper, as it is now the view method that will get stored for
        #   the app's routes
        self.store_auth_info(wrapper)
        return wrapper

    def on_authentication_failure(self):
        if self._on_authentication_failure:
            self._on_authentication_failure()

        # redirect if app's login manager requires it
        if flask.current_app.auth_manager.login_authenticator.authentication_failure_redirect:
            redirect_resp = flask.current_app.login_manager.unauthorized()
            flask.abort(redirect_resp)
        else:
            flask.abort(401)

    def check_auth(self, instance=None):
        methods_excluded = flask.current_app.config.get('KEGAUTH_HTTP_METHODS_EXCLUDED')
        if self.http_methods_excluded is not None:
            methods_excluded = self.http_methods_excluded
        if flask.request.method in methods_excluded:
            return

        # if flask_login has an authenticated user in session, that's who we want
        if flask_login.current_user.is_authenticated:
            return

        # no user in session right now, so we need to run request loaders to see if any match
        user = None
        for loader in flask.current_app.auth_manager.request_loaders.values():
            user = loader.get_authenticated_user()
            if user:
                break

        if not user or not user.is_authenticated:
            if instance and callable(getattr(instance, 'on_authentication_failure', None)):
                instance.on_authentication_failure()
            self.on_authentication_failure()


class RequiresPermissions(RequiresUser):
    """ Require a user to be conditionally authorized before proceeding to decorated target. May be
        used as a class decorator or method decorator.

        Usage: @requires_permissions(condition)

        Note: if using along with a route decorator (e.g. Blueprint.route), requires_permissions
            should be the closest decorator to the method

        Examples:
        - @requires_permissions(('token1', 'token2'))
        - @requires_permissions(has_any('token1', 'token2'))
        - @requires_permissions(has_all('token1', 'token2'))
        - @requires_permissions(has_all(has_any('token1', 'token2'), 'token3'))
        - @requires_permissions(custom_authorization_callable that takes user arg)
        - @requires_permissions('token1', on_authorization_failure=lambda: flask.abort(404))
    """
    def __init__(self, condition, on_authentication_failure=None, on_authorization_failure=None,
                 http_methods_excluded=None):
        super(RequiresPermissions, self).__init__(
            on_authentication_failure=on_authentication_failure,
            http_methods_excluded=http_methods_excluded,
        )
        self.condition = condition
        self._on_authorization_failure = on_authorization_failure

    def store_auth_info(self, obj):
        super(RequiresPermissions, self).store_auth_info(obj)
        condition = self.condition
        if callable(condition) and isinstance(obj, type):
            # When applying to a class (usually a class view or a blueprint), we have to explicitly
            # wrap callables as static. Otherwise, when the obj class is instantiated, the function
            # will become bound, and we'll get parameter count exceptions.
            condition = staticmethod(condition)
        obj.__keg_auth_requires_permissions__ = condition

    def on_authorization_failure(self):
        if self._on_authorization_failure:
            self._on_authorization_failure()
        flask.abort(403)

    def check_auth(self, instance=None):
        super(RequiresPermissions, self).check_auth(instance=instance)

        user = flask_login.current_user
        if self.condition and not model_utils.has_permissions(self.condition, user):
            if instance and callable(getattr(instance, 'on_authorization_failure', None)):
                instance.on_authorization_failure()
            self.on_authorization_failure()


def requires_user(arg=None, *args, **kwargs):
    """ Require a user to be authenticated before proceeding to decorated target. May be used as
        a class decorator or method decorator.

        Usage: @requires_user OR @requires_user() (both usage forms are identical)

        Parameters:
            on_authentication_failure: method called on authentication failures. If one is not
                specified, behavior is derived from login manager (redirect or 401)
            on_authorization_failure: method called on authorization failures. If one is not
                specified, response will be 403
    """
    if arg is None:
        return RequiresUser(*args, **kwargs)
    if inspect.isclass(arg):
        if issubclass(arg, flask.Blueprint):
            return RequiresUser().decorate_blueprint(arg)
        return RequiresUser().decorate_class(arg)  # pragma: no cover
    return RequiresUser().decorate_function(arg)


requires_permissions = RequiresPermissions



import flask
from markupsafe import Markup


def render_jinja(source, **kwargs):
    template = flask.current_app.jinja_env.from_string(source)
    return Markup(template.render(**kwargs))


def link_to(label, url, **kwargs):
    return render_jinja(
        '<a href="{{url}}" {{- attrs|html_attributes }}>{{label}}</a>',
        url=url,
        attrs=kwargs,
        label=label
    )



# Using unicode_literals instead of adding 'u' prefix to all stings that go to SA.
from __future__ import unicode_literals

import pytest
from flask import current_app
from mock import mock
from pyquery import PyQuery
from werkzeug.datastructures import MultiDict

from keg_auth import forms
import keg_auth_ta.model.entities as ents
from keg_auth.libs.authenticators import PasswordPolicy


class FormBase(object):
    form_cls = None

    def ok_data(self, **kwargs):
        return kwargs

    def make_form(self, **kwargs):
        form_cls = kwargs.pop('form_cls', None) or self.form_cls
        obj = kwargs.pop('obj', None)
        data = MultiDict(self.ok_data(**kwargs))
        return form_cls(data, obj=obj)

    def assert_valid(self, **kwargs):
        form = self.make_form(**kwargs)
        assert form.validate()
        return form

    def assert_not_valid(self, **kwargs):
        form = self.make_form(**kwargs)
        assert not form.validate()
        return form

    def test_ok(self):
        self.assert_valid()


@mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False)
class TestLogin(FormBase):
    form_cls = forms.login_form()

    def ok_data(self, **kwargs):
        data = {
            'login_id': 'foo@example.com',
            'password': 'password123',
        }
        data.update(kwargs)
        return data

    def test_required(self):
        form = self.assert_not_valid(login_id='', password='')
        msg = ['This field is required.']
        assert form.login_id.errors == msg
        assert form.password.errors == msg

    def test_valid_email(self):
        form = self.assert_not_valid(login_id='foo')
        assert form.login_id.errors == ['Invalid email address.']
        assert form.login_id.label.text == 'Email'

    def test_no_email_validation(self):
        with mock.patch('keg_auth_ta.extensions.auth_entity_registry._user_cls', ents.UserNoEmail):
            form_cls = forms.login_form()
            form = self.assert_valid(form_cls=form_cls, login_id='foo')
            assert form.login_id.label.text == 'User ID'


@mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False)
class TestForgotPassword(FormBase):
    form_cls = forms.ForgotPassword

    def ok_data(self, **kwargs):
        data = {
            'email': 'foo@example.com',
        }
        data.update(kwargs)
        return data

    def test_required(self):
        form = self.assert_not_valid(email='')
        msg = ['This field is required.']
        assert form.email.errors == msg

    def test_valid_email(self):
        form = self.assert_not_valid(email='foo')
        assert form.email.errors == ['Invalid email address.']


@mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False)
@mock.patch.object(current_app.auth_manager, 'password_policy_cls', new=PasswordPolicy)
class TestSetPassword(FormBase):
    def setup_method(self, _):
        ents.User.delete_cascaded()
        self.user = ents.User.testing_create()

    def ok_data(self, **kwargs):
        data = {
            'password': 'password123!',
            'confirm': 'password123!',
        }
        data.update(kwargs)
        return data

    def form_cls(self, *args, **kwargs):
        kwargs.setdefault('user', self.user)
        return forms.SetPassword(*args, **kwargs)

    def test_required(self):
        form = self.assert_not_valid(password='')
        assert form.password.errors == ['This field is required.']

    def test_valid_confirm(self):
        form = self.assert_not_valid(confirm='password1234')
        assert form.password.errors == ['Passwords must match']

    def test_length_validator(self):
        form = self.assert_not_valid(password='aBcDe1!', confirm='aBcDe1!')
        assert form.password.errors == ['Password must be at least 8 characters long']

        self.assert_valid(password='aBcDeF1!', confirm='aBcDeF1!')

    @pytest.mark.parametrize('pw', [
        'a' * 10,
        'A' * 10,
        '1' * 10,
        'aA' * 5,
        'a1' * 5,
        '1!' * 5,
    ])
    def test_char_set_validator_failures(self, pw):
        form = self.assert_not_valid(password=pw, confirm=pw)
        assert form.password.errors == [
            'Password must include at least 3 of lowercase letter, uppercase letter, number and/or symbol'  # noqa: E501
        ]

    @pytest.mark.parametrize('pw', [
        'aaaaaaaa1!',
        'aaaaaaaaA!',
        'aaaaaaaaA1',
        'AAAAAAAA1!',
    ])
    def test_char_set_validator_pass(self, pw):
        self.assert_valid(password=pw, confirm=pw)

    @pytest.mark.parametrize('pw,email', [
        ('1!bob!1234', 'bob@example.com'),
        ('BoB123456!', 'bOb@example.com'),
    ])
    def test_username_validator_failures(self, pw, email):
        self.user.email = email
        form = self.assert_not_valid(password=pw, confirm=pw)
        assert form.password.errors == ['Password may not contain username']

    @pytest.mark.parametrize('pw,email', [
        ('1!b0b!1234', 'bob@example.com'),
        ('B0B123456!', 'bOb@example.com'),
    ])
    def test_username_validator_pass(self, pw, email):
        self.user.email = email
        self.assert_valid(password=pw, confirm=pw)


@mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False)
class TestUser(FormBase):
    form_cls = forms.user_form({'KEGAUTH_EMAIL_OPS_ENABLED': True},
                               allow_superuser=False, endpoint='auth.user:edit')

    @classmethod
    def setup_class(cls):
        ents.Permission.delete_cascaded()
        ents.Group.delete_cascaded()
        ents.Bundle.delete_cascaded()

        cls.perms = [ents.Permission.testing_create() for _ in range(3)]
        cls.groups = [ents.Group.testing_create() for _ in range(3)]
        cls.bundles = [ents.Bundle.testing_create() for _ in range(3)]

    def setup(self):
        ents.User.delete_cascaded()

    def ok_data(self, **kwargs):
        data = {
            'email': 'foo@example.com',
            'is_enabled': 'true',
            'disabled_utc': '2019-01-01',
            'permission_ids': [str(self.perms[0].id), str(self.perms[1].id)],
            'group_ids': [str(self.groups[0].id), str(self.groups[1].id)],
            'bundle_ids': [str(self.bundles[0].id), str(self.bundles[1].id)],
        }
        data.update(kwargs)
        return data

    def test_required(self):
        form = self.assert_not_valid(email='')
        assert form.email.errors == ['This field is required.']

    def test_valid_email(self):
        form = self.assert_not_valid(email='foo')
        assert form.email.errors == ['Invalid email address.']

    def test_superuser_available(self):
        form = self.make_form()
        assert not hasattr(form, 'is_superuser')

        form = forms.user_form({'KEGAUTH_EMAIL_OPS_ENABLED': True},
                               allow_superuser=True, endpoint='auth.user:edit')
        assert hasattr(form, 'is_superuser')

    def test_alternate_ident_field(self):
        with mock.patch('keg_auth_ta.extensions.auth_entity_registry._user_cls', ents.UserNoEmail):
            form_cls = forms.user_form({'KEGAUTH_EMAIL_OPS_ENABLED': True},
                                       allow_superuser=False, endpoint='auth.user:edit')
            assert hasattr(form_cls, 'username')

    def test_send_welcome_present(self):
        form = self.make_form()
        assert form.send_welcome

    def test_send_welcome_absent(self):
        user = ents.User.testing_create()
        form = self.make_form(obj=user)
        assert not form.send_welcome

    def test_no_email(self):
        with mock.patch('keg_auth_ta.extensions.auth_entity_registry._user_cls', ents.UserNoEmail):
            form_cls = forms.user_form({'KEGAUTH_EMAIL_OPS_ENABLED': False},
                                       allow_superuser=False, endpoint='auth.user:edit')
            assert hasattr(form_cls, 'username')
            assert not hasattr(form_cls, 'email')
            assert hasattr(form_cls, 'reset_password')
            assert hasattr(form_cls, 'confirm')

            form = self.assert_not_valid(form_cls=form_cls, reset_password='xyz', confirm='abc')
            assert form.reset_password.errors == ['Passwords must match']
            self.assert_valid(form_cls=form_cls, username='foobar', reset_password='xyz',
                              confirm='xyz')

    def test_multi_select(self):
        form = self.assert_valid()
        assert form.get_selected_permissions() == self.perms[:2]
        assert form.get_selected_groups() == self.groups[:2]
        assert form.get_selected_bundles() == self.bundles[:2]

        form = self.assert_valid(permission_ids=[], group_ids=[], bundle_ids=[])
        assert form.get_selected_permissions() == []
        assert form.get_selected_groups() == []
        assert form.get_selected_bundles() == []

    def test_unique(self):
        usr = ents.User.testing_create(email='foo@example.com')

        form = self.assert_not_valid()
        error = PyQuery(form.email.errors[1])
        assert 'This value must be unique' in error.text()
        assert error('a').attr('href').endswith('/users/{}/edit'.format(usr.id))
        assert error('a').text() == 'foo@example.com'

        self.assert_valid(obj=usr)

    def test_unique_alternate_ident_field(self):
        with mock.patch('keg_auth_ta.extensions.auth_entity_registry._user_cls', ents.UserNoEmail):
            form_cls = forms.user_form({'KEGAUTH_EMAIL_OPS_ENABLED': True},
                                       allow_superuser=False, endpoint='auth.user:edit')
            usr = ents.UserNoEmail.testing_create(username='foobar')

            form = self.assert_not_valid(form_cls=form_cls, username='foobar')
            error = PyQuery(form.username.errors[1])
            assert 'This value must be unique' in error.text()
            assert error('a').attr('href').endswith('/users/{}/edit'.format(usr.id))
            assert error('a').text() == 'foobar'

            self.assert_valid(form_cls=form_cls, obj=usr, username='foobar')


@mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False)
class TestGroup(FormBase):
    form_cls = forms.group_form(endpoint='auth.group:edit')

    @classmethod
    def setup_class(cls):
        ents.Permission.delete_cascaded()
        ents.Bundle.delete_cascaded()

        cls.perms = [ents.Permission.testing_create() for _ in range(3)]
        cls.bundles = [ents.Bundle.testing_create() for _ in range(3)]

    def ok_data(self, **kwargs):
        data = {
            'name': 'some-group',
            'permission_ids': [str(self.perms[0].id), str(self.perms[1].id)],
            'bundle_ids': [str(self.bundles[0].id), str(self.bundles[1].id)],
        }
        data.update(kwargs)
        return data

    def test_required(self):
        form = self.assert_not_valid(name='')
        assert form.name.errors == ['This field is required.']

    def test_multi_select(self):
        form = self.assert_valid()
        assert form.get_selected_permissions() == self.perms[:2]
        assert form.get_selected_bundles() == self.bundles[:2]

        form = self.assert_valid(permission_ids=[], bundle_ids=[])
        assert form.get_selected_permissions() == []
        assert form.get_selected_bundles() == []

    def test_unique(self):
        obj = ents.Group.testing_create(name='some-group')

        form = self.assert_not_valid()
        print(form.name.errors[0])
        error = PyQuery(form.name.errors[0])
        assert 'Already exists.' in error.text()

        self.assert_valid(obj=obj)


@mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False)
class TestBundle(FormBase):
    form_cls = forms.bundle_form(endpoint='auth.bundle:edit')

    @classmethod
    def setup_class(cls):
        ents.Permission.delete_cascaded()

        cls.perms = [ents.Permission.testing_create() for _ in range(3)]

    def ok_data(self, **kwargs):
        data = {
            'name': 'some-bundle',
            'permission_ids': [str(self.perms[0].id), str(self.perms[1].id)],
        }
        data.update(kwargs)
        return data

    def test_required(self):
        form = self.assert_not_valid(name='')
        assert form.name.errors == ['This field is required.']

    def test_multi_select(self):
        form = self.assert_valid()
        assert form.get_selected_permissions() == self.perms[:2]

        form = self.assert_valid(permission_ids=[])
        assert form.get_selected_permissions() == []

    def test_unique(self):
        obj = ents.Bundle.testing_create(name='some-bundle')

        form = self.assert_not_valid()
        print(form.name.errors[0])
        error = PyQuery(form.name.errors[0])
        assert 'Already exists.' in error.text()

        self.assert_valid(obj=obj)



# flake8: noqa
from keg_auth.core import AuthManager
from keg_auth.libs import get_current_user
from keg_auth.libs.authenticators import (
    JwtRequestLoader,
    KegAuthenticator,
    LdapAuthenticator,
    TokenRequestLoader,
    PasswordPolicy,
    PasswordPolicyError,
)
from keg_auth.libs.decorators import requires_permissions, requires_user
from keg_auth.libs.navigation import NavItem, NavURL
from keg_auth.mail import AuthMailManager
from keg_auth.model import (
    UserMixin,
    UserEmailMixin,
    UserTokenMixin,
    AttemptMixin,
    PermissionMixin,
    GroupMixin,
    BundleMixin,
    initialize_mappings
)
from keg_auth.model.entity_registry import EntityRegistry as AuthEntityRegistry
from keg_auth.model.utils import (
    has_permissions,
    has_all,
    has_any,
)
from keg_auth.version import VERSION as __version__
from keg_auth.views import (
    CrudView,
    make_blueprint,
)



from __future__ import absolute_import

from flask_mail import Mail
from flask_wtf.csrf import CSRFProtect
from keg_auth import AuthManager, AuthMailManager, JwtRequestLoader, AuthEntityRegistry
import webgrid
from webgrid.flask import WebGrid as GridManager


class Grid(webgrid.BaseGrid):
    manager = GridManager()
    session_on = True


csrf = CSRFProtect()

permissions = (
    'auth-manage',
    'permission1',
    'permission2',
)

auth_entity_registry = AuthEntityRegistry()
mail_ext = Mail()
_endpoints = {'after-login': 'public.home'}
auth_mail_manager = AuthMailManager(mail_ext)
auth_manager = AuthManager(mail_manager=auth_mail_manager, endpoints=_endpoints, grid_cls=Grid,
                           request_loaders=[JwtRequestLoader], entity_registry=auth_entity_registry,
                           permissions=permissions)



# Using unicode_literals instead of adding 'u' prefix to all stings that go to SA.
from __future__ import unicode_literals
import base64
import string
import arrow
import flask
from keg.db import db
import pytest
from freezegun import freeze_time
import sqlalchemy as sa
import bcrypt

from keg_auth.model import InvalidToken, entity_registry, utils
from keg_auth_ta.model import entities as ents
from keg_auth.testing import with_crypto_context
import mock


class TestUserTokenMixin(object):
    def setup(self):
        ents.UserWithToken.delete_cascaded()

    @with_crypto_context(ents.UserWithToken.token)
    def test_token_storage_with_real_bcrypt(self):
        raw_token = b'a' * 32
        uwt = ents.UserWithToken.testing_create(token=raw_token)
        assert bcrypt.checkpw(raw_token, uwt.token.hash)

    @with_crypto_context(ents.UserWithToken.token)
    def test_reset_auth_token(self):
        original_token = b'a' * 32
        uwt = ents.UserWithToken.testing_create(token=original_token)

        assert bcrypt.checkpw(original_token, uwt.token.hash)
        new_token = uwt.reset_auth_token()

        assert bcrypt.checkpw(new_token.encode(), uwt.token.hash)
        assert not bcrypt.checkpw(original_token, uwt.token.hash)

    def test_generate_auth_token(self):
        assert len(ents.UserWithToken.generate_raw_auth_token(length=1)) == 1
        assert len(ents.UserWithToken.generate_raw_auth_token()) == 32
        assert len(ents.UserWithToken.generate_raw_auth_token(entropy='secure')) == 32
        assert len(ents.UserWithToken.generate_raw_auth_token(length=None, entropy='secure')) == 11
        assert (string.ascii_uppercase not in
                ents.UserWithToken.generate_raw_auth_token(charset='hex'))

    @pytest.mark.parametrize('raw,test,result', [
        ('123', '123', True),
        ('123', 'abc', False),
        ('a' * 400, 'abc', False),
        (b'abc', b'abc', True),
        (b'abc', 'abc', True),
        ('abc', b'abc', True),
        (None, b'abc', False),
        ('abc', None, False),
    ])
    def test_verify_token(self, raw, test, result):
        uwt = ents.UserWithToken.testing_create(token=raw)
        assert uwt.verify_token(test) is result

    def test_generate_api_token(self):
        u1 = ents.UserWithToken.testing_create()
        raw_token = u1.generate_api_token()

        raw_email, token = raw_token.split('.')
        real_email = base64.urlsafe_b64decode(raw_email.encode()).decode()

        assert (real_email, token) == (u1.email, u1.token.hash.decode())

    def test_get_user_for_api_token_happy_path(self):
        u1 = ents.UserWithToken.testing_create(email='test@test.com', token='1234')
        ents.UserWithToken.testing_create(email='test-2@test.com', token='5678')

        u1_api_token = u1.generate_api_token('1234')
        assert ents.UserWithToken.get_user_for_api_token(u1_api_token).id == u1.id

    def test_get_user_for_api_token_wrong_token(self):
        u1 = ents.UserWithToken.testing_create(email='test@test.com', token='1234')
        ents.UserWithToken.testing_create(email='test-2@test.com', token='5678')

        u1_api_token = u1.generate_api_token('1234')
        u1_raw_email, u1_raw_token = u1_api_token.split('.')
        u1_fake_token = '{email}.faketoken'.format(email=u1_raw_email)

        assert ents.UserWithToken.get_user_for_api_token(u1_fake_token) is None

    @pytest.mark.parametrize('token', [
        None,  # Bad
        'test@testcom',  # No period
        'tes.t@te.stcom',  # three periods
        b'test@thing.token',  # not b64encoded
    ])
    def test_get_user_for_api_token_bad_token(self, token):
        assert ents.UserWithToken.get_user_for_api_token(token) is None


class TestUser(object):
    def setup(self):
        ents.User.delete_cascaded()
        ents.Permission.delete_cascaded()

    def test_email_case_insensitive(self):
        ents.User.testing_create(email='foo@BAR.com')

        assert ents.User.get_by(email='foo@bar.com')

    def test_is_verified_default(self):
        # testing_create() overrides the is_enabled default to make testing easier.  So, make sure
        # that we have set enabled to False when not used in a testing environment.
        user = ents.User.add(email='foo', password='bar')
        assert not user.is_verified
        assert user.password == 'bar'

    @pytest.mark.parametrize('is_enabled, is_verified, disabled_utc, is_active', [
        (True, True, None, True),
        (True, True, arrow.utcnow().shift(minutes=+5), True),
        (True, False, None, False),
        (False, True, None, False),
        (True, True, arrow.utcnow().shift(minutes=-5), False),
    ])
    def test_is_active_python_attribute(self, is_enabled, is_verified, disabled_utc, is_active):
        user = ents.User.testing_create(
            is_verified=is_verified,
            is_enabled=is_enabled,
            disabled_utc=disabled_utc,
        )
        assert user.is_active == is_active

    @pytest.mark.parametrize('is_enabled, is_verified, disabled_utc, is_active', [
        (True, True, None, True),
        (True, True, arrow.utcnow().shift(minutes=+5), True),
        (True, False, None, False),
        (False, True, None, False),
        (True, True, arrow.utcnow().shift(minutes=-5), False),
    ])
    def test_is_active_sql_expression(self, is_enabled, is_verified, disabled_utc, is_active):
        ents.User.testing_create(
            email='email',
            is_verified=is_verified,
            is_enabled=is_enabled,
            disabled_utc=disabled_utc,
        )
        assert ents.User.query.filter_by(email='email', is_active=is_active).one()

    def test_token_validation(self):
        user = ents.User.testing_create()

        assert not hasattr(user, '_token_plain')
        assert not user.token_verify(None)

        token = user.token_generate()
        assert token
        assert not user.token_verify('foo')
        assert user.token_verify(token)
        assert user.token_verify(user._token_plain)

    def test_token_salt_info_changed(self):
        def check_field(field, new_value):
            user = ents.User.testing_create(last_login_utc=None)
            token = user.token_generate()
            setattr(user, field, new_value)
            db.session.flush()
            db.session.commit()
            db.session.refresh(user)
            assert not user.token_verify(token)

        check_field('email', 'foobar')
        check_field('is_enabled', False)
        check_field('is_verified', False)
        check_field('password', 'foobar')
        check_field('last_login_utc', arrow.utcnow())

    def test_token_expiration(self):
        user = ents.User.add(email='foo', password='bar')

        with mock.patch.dict(flask.current_app.config, KEGAUTH_TOKEN_EXPIRE_MINS=10):
            token = user.token_generate()
            now = arrow.get()
            assert user.token_verify(token)

            plus_9_58 = now.shift(minutes=9, seconds=58).datetime
            with freeze_time(plus_9_58):
                assert user.token_verify(token)
            plus_10_01 = now.shift(minutes=10, seconds=1).datetime
            with freeze_time(plus_10_01):
                assert not user.token_verify(token)

    def test_change_password(self):
        user = ents.User.testing_create(is_verified=False)
        token = user.token_generate()
        user.change_password(token, 'abc123')
        assert not user.token_verify(token)
        assert user.password == 'abc123'
        assert user.is_verified

    def test_change_password_invalid_token(self):
        user = ents.User.testing_create(is_verified=False)
        with pytest.raises(InvalidToken):
            user.change_password('bad-token', 'abc123')

    def test_permissions_mapping(self):
        perm1 = ents.Permission.testing_create()
        perm2 = ents.Permission.testing_create()
        perm3 = ents.Permission.testing_create()
        perm4 = ents.Permission.testing_create()
        perm5 = ents.Permission.testing_create()

        bundle1 = ents.Bundle.testing_create()
        bundle2 = ents.Bundle.testing_create()
        bundle3 = ents.Bundle.testing_create()

        group1 = ents.Group.testing_create()
        group2 = ents.Group.testing_create()
        group3 = ents.Group.testing_create()

        user1 = ents.User.testing_create()
        user2 = ents.User.testing_create()

        # Directly assigned
        user1.permissions = [perm1]

        # Assigned via user bundle
        bundle1.permissions = [perm2]
        user1.bundles = [bundle1]

        # Assigned via group
        group1.permissions = [perm3]

        # Assigned via group bundle
        bundle2.permissions = [perm4]
        group1.bundles = [bundle2]
        user1.groups = [group1, group2]

        assert user1.get_all_permissions() == {perm1, perm2, perm3, perm4}
        assert user2.get_all_permissions() == set()

        user2.permissions = [perm1, perm2]
        group3.permissions = [perm2, perm3]
        bundle3.permissions = [perm1, perm5]
        group3.bundles = [bundle3]
        user2.groups = [group3]

        assert user1.get_all_permissions() == {perm1, perm2, perm3, perm4}
        assert user2.get_all_permissions() == {perm1, perm2, perm3, perm5}

        user1.is_superuser = True
        assert user1.get_all_permissions() == {perm1, perm2, perm3, perm4, perm5}

    def test_get_all_permission_tokens(self):
        ents.Permission.delete_cascaded()
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')
        perm3 = ents.Permission.testing_create(token='perm-3')

        user = ents.User.testing_create(permissions=[perm1, perm2, perm3])

        assert user.get_all_permission_tokens() == {'perm-1', 'perm-2', 'perm-3'}

    def test_get_all_permission_tokens_cached(self):
        ents.Permission.delete_cascaded()
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')
        perm3 = ents.Permission.testing_create(token='perm-3')

        user = ents.User.testing_create(permissions=[perm1, perm2])
        # trigger the cache storage
        assert user.get_all_permission_tokens() == {'perm-1', 'perm-2'}

        # permissions in cache go stale
        user.permissions = [perm1, perm2, perm3]
        db.session.commit()

        assert user.get_all_permission_tokens() == {'perm-1', 'perm-2'}

        # reset the cache
        delattr(user, '_permission_cache')

        assert user.get_all_permission_tokens() == {'perm-1', 'perm-2', 'perm-3'}

    def test_has_all_permissions(self):
        ents.Permission.delete_cascaded()
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')
        ents.Permission.testing_create(token='perm-3')

        user = ents.User.testing_create(permissions=[perm1, perm2])

        assert user.has_all_permissions('perm-1', 'perm-2') is True
        assert user.has_all_permissions('perm-1', 'perm-3') is False
        assert user.has_all_permissions('perm-1') is True
        assert user.has_all_permissions('perm-3') is False

    def test_has_any_permission(self):
        ents.Permission.delete_cascaded()
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')
        ents.Permission.testing_create(token='perm-3')

        user = ents.User.testing_create(permissions=[perm1, perm2])

        assert user.has_any_permission('perm-1', 'perm-2') is True
        assert user.has_any_permission('perm-1', 'perm-3') is True
        assert user.has_any_permission('perm-1') is True
        assert user.has_any_permission('perm-3') is False

    def test_superuser_update_resets_session_key(self):
        user = ents.User.testing_create(is_superuser=True)
        original_session_key = user.session_key

        ents.User.edit(user.id, is_superuser=False)
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_enabled_update_resets_session_key(self):
        user = ents.User.testing_create(is_enabled=True)
        original_session_key = user.session_key

        ents.User.edit(user.id, is_enabled=False)
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_permission_update_resets_session_key(self):
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')

        user = ents.User.testing_create(permissions=[perm1])
        original_session_key = user.session_key

        ents.User.edit(user.id, permissions=[perm2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_group_update_resets_session_key(self):
        group1 = ents.Group.testing_create(name='group-1')
        group2 = ents.Group.testing_create(name='group-2')

        user = ents.User.testing_create(groups=[group1])
        original_session_key = user.session_key

        ents.User.edit(user.id, groups=[group2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_bundle_update_resets_session_key(self):
        bundle1 = ents.Bundle.testing_create(name='bundle-1')
        bundle2 = ents.Bundle.testing_create(name='bundle-2')

        user = ents.User.testing_create(bundles=[bundle1])
        original_session_key = user.session_key

        ents.User.edit(user.id, bundles=[bundle2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_non_permission_update_does_not_reset_session_key(self):
        user = ents.User.testing_create()
        original_session_key = user.session_key

        ents.User.edit(user.id, email='foo@bar.baz')
        db.session.expire(user)
        assert user.session_key == original_session_key


class TestUserNoEmail(object):
    def setup(self):
        ents.UserNoEmail.delete_cascaded()

    @pytest.mark.parametrize('is_enabled, disabled_utc, is_active', [
        (True, None, True),
        (True, arrow.utcnow().shift(minutes=+5), True),
        (False, None, False),
        (True, arrow.utcnow().shift(minutes=-5), False),
    ])
    def test_is_active_python_attribute(self, is_enabled, disabled_utc, is_active):
        user = ents.UserNoEmail.testing_create(
            is_enabled=is_enabled,
            disabled_utc=disabled_utc,
        )
        assert user.is_active == is_active

    @pytest.mark.parametrize('is_enabled, disabled_utc, is_active', [
        (True, None, True),
        (True, arrow.utcnow().shift(minutes=+5), True),
        (False, None, False),
        (True, arrow.utcnow().shift(minutes=-5), False),
    ])
    def test_is_active_sql_expression(self, is_enabled, disabled_utc, is_active):
        ents.UserNoEmail.testing_create(
            username='name',
            is_enabled=is_enabled,
            disabled_utc=disabled_utc,
        )

        assert ents.UserNoEmail.query.filter_by(username='name', is_active=is_active).one()


class TestPermission(object):
    def setup(self):
        ents.Permission.delete_cascaded()

    def test_token_unique(self):
        ents.Permission.testing_create(token='some-permission')
        with pytest.raises(sa.exc.IntegrityError) as exc:
            # use `add` here instead of `testing_create`, because it is more helpful for the
            #   `testing_create` method to return the existing permission if there is a match
            ents.Permission.add(token='some-permission')

        assert 'unique' in str(exc.value).lower()


class TestBundle(object):
    def setup(self):
        ents.Bundle.delete_cascaded()

    def test_name_unique(self):
        ents.Bundle.testing_create(name='Bundle 1')
        with pytest.raises(sa.exc.IntegrityError) as exc:
            ents.Bundle.testing_create(name='Bundle 1')

        assert 'unique' in str(exc.value).lower()

    def test_permission_update_resets_user_session_keys(self):
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')

        user = ents.User.testing_create()
        bundle = ents.Bundle.testing_create(permissions=[perm1], users=[user])
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, permissions=[perm2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_permission_update_resets_group_user_session_keys(self):
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')

        user = ents.User.testing_create()
        group = ents.Group.testing_create(users=[user])
        bundle = ents.Bundle.testing_create(permissions=[perm1], groups=[group])
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, permissions=[perm2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_group_addition_resets_user_session_keys(self):
        user = ents.User.testing_create()
        group = ents.Group.testing_create(users=[user])
        bundle = ents.Bundle.testing_create()
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, groups=[group])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_group_removal_resets_user_session_keys(self):
        user = ents.User.testing_create()
        group = ents.Group.testing_create(users=[user])
        bundle = ents.Bundle.testing_create(groups=[group])
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, groups=[])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_user_addition_resets_user_session_keys(self):
        user = ents.User.testing_create()
        bundle = ents.Bundle.testing_create()
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, users=[user])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_user_removal_resets_user_session_keys(self):
        user = ents.User.testing_create()
        bundle = ents.Bundle.testing_create(users=[user])
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, users=[])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_bundle_removal_resets_user_session_keys(self):
        user = ents.User.testing_create()
        bundle = ents.Bundle.testing_create(users=[user])
        original_session_key = user.session_key
        ents.Bundle.delete(bundle.id)

        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_bundle_removal_resets_group_session_keys(self):
        user = ents.User.testing_create()
        group = ents.Group.testing_create(users=[user])
        bundle = ents.Bundle.testing_create(groups=[group])
        original_session_key = user.session_key
        ents.Bundle.delete(bundle.id)

        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_non_permission_update_does_not_reset_user_session_keys(self):
        user = ents.User.testing_create()
        bundle = ents.Bundle.testing_create(users=[user])
        original_session_key = user.session_key

        ents.Bundle.edit(bundle.id, name='foo')
        db.session.expire(user)
        assert user.session_key == original_session_key


class TestGroup(object):
    def setup(self):
        ents.Group.delete_cascaded()

    def test_name_unique(self):
        ents.Group.testing_create(name='Group 1')
        with pytest.raises(sa.exc.IntegrityError) as exc:
            ents.Group.testing_create(name='Group 1')

        assert 'unique' in str(exc.value).lower()

    def test_get_all_permissions(self):
        perm1 = ents.Permission.testing_create()
        perm2 = ents.Permission.testing_create()
        perm3 = ents.Permission.testing_create()

        bundle = ents.Bundle.testing_create()

        group1 = ents.Group.testing_create()
        group2 = ents.Group.testing_create()

        # Assigned directly
        group1.permissions = [perm1]

        # Assigned via bundle
        bundle.permissions = [perm2]
        group1.bundles = [bundle]

        assert group1.get_all_permissions() == {perm1, perm2}
        assert group2.get_all_permissions() == set()

        group2.bundles = [bundle]
        group2.permissions = [perm2, perm3]

        assert group1.get_all_permissions() == {perm1, perm2}
        assert group2.get_all_permissions() == {perm2, perm3}

    def test_permission_update_resets_user_session_keys(self):
        perm1 = ents.Permission.testing_create(token='perm-1')
        perm2 = ents.Permission.testing_create(token='perm-2')

        user = ents.User.testing_create()
        original_session_key = user.session_key
        group = ents.Group.testing_create(permissions=[perm1], users=[user])

        ents.Group.edit(group.id, permissions=[perm2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_bundle_update_resets_user_session_keys(self):
        bundle1 = ents.Bundle.testing_create(name='bundle-1')
        bundle2 = ents.Bundle.testing_create(name='bundle-2')

        user = ents.User.testing_create()
        original_session_key = user.session_key
        group = ents.Group.testing_create(bundles=[bundle1], users=[user])

        ents.Group.edit(group.id, bundles=[bundle2])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_user_addition_resets_user_session_keys(self):
        user = ents.User.testing_create()
        original_session_key = user.session_key
        group = ents.Group.testing_create()

        ents.Group.edit(group.id, users=[user])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_user_removal_resets_user_session_keys(self):
        user = ents.User.testing_create()
        original_session_key = user.session_key
        group = ents.Group.testing_create(users=[user])

        ents.Group.edit(group.id, users=[])
        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_group_removal_resets_user_session_keys(self):
        user = ents.User.testing_create()
        group = ents.Group.testing_create(users=[user])
        original_session_key = user.session_key
        ents.Group.delete(group.id)

        db.session.expire(user)
        assert user.session_key != original_session_key

    def test_non_permission_update_does_not_reset_user_session_keys(self):
        user = ents.User.testing_create()
        group = ents.Group.testing_create(users=[user])
        original_session_key = user.session_key

        ents.Group.edit(group.id, name='foo')
        db.session.expire(user)
        assert user.session_key == original_session_key


class TestEntityRegistry(object):
    def test_bad_type(self):
        registry = entity_registry.EntityRegistry()
        with pytest.raises(entity_registry.RegistryError):
            registry.get_entity_cls('foo')

    def test_entity_not_defined(self):
        registry = entity_registry.EntityRegistry()
        with pytest.raises(entity_registry.RegistryError):
            registry.get_entity_cls('user')

    def test_register_entities(self):
        registry = entity_registry.EntityRegistry()

        @registry.register_user
        class TestingUser(object):
            pass

        @registry.register_permission
        class TestingPermission(object):
            pass

        @registry.register_bundle
        class TestingBundle(object):
            pass

        @registry.register_group
        class TestingGroup(object):
            pass

        assert registry.user_cls is TestingUser
        assert registry.permission_cls is TestingPermission
        assert registry.bundle_cls is TestingBundle
        assert registry.group_cls is TestingGroup

    def test_duplicate_registration(self):
        registry = entity_registry.EntityRegistry()

        @registry.register_user
        class TestingUser1(object):
            pass

        with pytest.raises(entity_registry.RegistryError) as exc:
            @registry.register_user
            class TestingUser2(object):
                pass

        assert str(exc.value) == 'Entity class already registered for user'

    def test_register_unknown_type(self):
        registry = entity_registry.EntityRegistry()

        class Foo(object):
            pass

        with pytest.raises(entity_registry.RegistryError) as exc:
            registry.register_entity('foo', Foo)

        assert str(exc.value) == 'Attempting to register unknown type foo'

    def test_register_nonclass(self):
        registry = entity_registry.EntityRegistry()

        with pytest.raises(entity_registry.RegistryError) as exc:
            @registry.register_user
            def testing_user():
                pass

        assert str(exc.value) == 'Entity must be a class'

        with pytest.raises(entity_registry.RegistryError) as exc:
            registry.register_user(ents.User.testing_create())

        assert str(exc.value) == 'Entity must be a class'

    def test_is_registered(self):
        registry = entity_registry.EntityRegistry()

        @registry.register_user
        class TestingUser(object):
            pass

        @registry.register_permission
        class TestingPermission(object):
            pass

        assert registry.is_registered('user') is True
        assert registry.is_registered('permission') is True
        assert registry.is_registered('bundle') is False
        assert registry.is_registered('group') is False


class TestPermissionsConditions:
    def setup(self):
        ents.Permission.delete_cascaded()
        ents.User.delete_cascaded()

    def test_no_conditions(self):
        with pytest.raises(ValueError):
            utils.PermissionCondition()

    def test_simple_string(self):
        user = ents.User.testing_create(
            permissions=[ents.Permission.testing_create(token='perm1')]
        )
        ents.Permission.testing_create(token='perm2')

        assert utils.has_any('perm1').check(user) is True
        assert utils.has_all('perm1').check(user) is True

        assert utils.has_any('perm2').check(user) is False
        assert utils.has_all('perm2').check(user) is False

    def test_callable(self):
        user1 = ents.User.testing_create(email='foo@bar.com')
        user2 = ents.User.testing_create(email='abc@123.com')

        def func(usr):
            return usr.email.endswith('@bar.com')

        assert utils.has_any(func).check(user1) is True
        assert utils.has_all(func).check(user1) is True

        assert utils.has_any(func).check(user2) is False
        assert utils.has_all(func).check(user2) is False

    def test_all(self):
        user = ents.User.testing_create(
            permissions=[
                ents.Permission.testing_create(token='perm1'),
                ents.Permission.testing_create(token='perm2'),
                ents.Permission.testing_create(token='perm3'),
            ]
        )
        ents.Permission.testing_create(token='perm4')

        assert utils.has_all('perm1').check(user) is True
        assert utils.has_all('perm1', 'perm2').check(user) is True
        assert utils.has_all('perm1', 'perm2', 'perm3').check(user) is True

        assert utils.has_all('perm4').check(user) is False
        assert utils.has_all('perm1', 'perm4').check(user) is False
        assert utils.has_all('perm1', 'perm2', 'perm4').check(user) is False

    def test_any(self):
        user = ents.User.testing_create(
            permissions=[ents.Permission.testing_create(token='perm1')]
        )
        ents.Permission.testing_create(token='perm2'),
        ents.Permission.testing_create(token='perm3'),
        ents.Permission.testing_create(token='perm4')

        assert utils.has_any('perm1').check(user) is True
        assert utils.has_any('perm1', 'perm2').check(user) is True
        assert utils.has_any('perm1', 'perm2', 'perm3').check(user) is True

        assert utils.has_any('perm2').check(user) is False
        assert utils.has_any('perm2', 'perm3').check(user) is False
        assert utils.has_any('perm2', 'perm3', 'perm4').check(user) is False

    def test_nested(self):
        user = ents.User.testing_create(
            permissions=[
                ents.Permission.testing_create(token='perm1'),
                ents.Permission.testing_create(token='perm2'),
                ents.Permission.testing_create(token='perm3'),
            ]
        )
        ents.Permission.testing_create(token='perm4')

        condition = utils.has_any('perm4', utils.has_all('perm1', 'perm2'))
        assert condition.check(user) is True

        condition = utils.has_all(utils.has_any('perm1', 'perm2'), 'perm4')
        assert condition.check(user) is False

        condition = utils.has_all(utils.has_any('perm4', lambda _: True), 'perm1')
        assert condition.check(user) is True

        condition = utils.has_all(utils.has_any('perm4', lambda _: False), 'perm1')
        assert condition.check(user) is False


"""class TestPerformance(object):
    # check how long the SA events add to the process
    def test_performance(self):
        import time
        start = time.time()
        for _ in range(1000):
            ents.User.testing_create()
            ents.Group.testing_create()
            ents.Bundle.testing_create()
        assert False, time.time() - start
"""



from datetime import timedelta
from urllib.parse import urljoin, urlparse

import arrow
import flask
import flask_login
import sqlalchemy as sa
import string
import typing
import wtforms
from keg.db import db
from sqlalchemy_utils import EmailType

from keg_auth import forms
from keg_auth.extensions import flash, lazy_gettext as _
from keg_auth.model import get_username_key, get_username
from keg_auth.model.entity_registry import RegistryError

try:
    import flask_jwt_extended
except ImportError:
    pass  # pragma: no cover

try:
    import ldap
except ImportError:
    pass  # pragma: no cover


class UserNotFound(Exception):
    pass


class UserInactive(Exception):
    def __init__(self, user):
        self.user = user


class UserInvalidAuth(Exception):
    def __init__(self, user):
        self.user = user


class RequestLoader(object):
    """ Generic loader interface for determining if a user should be logged in"""

    def __init__(self, app):
        self.user_ent = app.auth_manager.entity_registry.user_cls

    @classmethod
    def get_identifier(cls):
        return cls.__name__.lower().replace('requestloader', '')


class LoginAuthenticator(object):
    """ Manages verification of users as well as relevant view-layer logic

        Relevant auth views (login, verification, resets, etc.) get passed through to responders
        on this layer, to process and render for the specific type of authentication happening.

        For example, a password authenticator will want a user/password login form, but other types
        like oauth may get a different form entirely (and handle resets differently, etc.).

        `responder_cls` is a key/value store for associating view keys with responder classes. If a
        view key is not present, we assume that view is not relevant to the authenticator, and the
        view itself will return 404.
    """

    authentication_failure_redirect = True
    responder_cls = {}

    def __init__(self, app):
        self.user_ent = app.auth_manager.entity_registry.user_cls
        self.responders = {}

        self.init_responders()

    def init_responders(self):
        for key, cls in self.responder_cls.items():
            self.responders[key] = cls(self)

    def get_responder(self, key):
        return self.responders.get(key)


class ViewResponder(object):
    """ View-layer logic wrapper for use in the Authenticator

        Responder should be combined with needed mixins for various functionality (forms,
        logins, etc.).

        Expected to have methods named for the request method (get, post, etc.)

        `template_name` is passed to `flask.render_template` by default
    """
    template_name = None

    def __init__(self, parent):
        self.template_args = {}
        self.parent = parent

    def assign(self, key, value):
        self.template_args[key] = value

    def render(self):
        return flask.render_template(self.template_name, **self.template_args)

    def __call__(self, *args, **kwargs):
        method_name = flask.request.method.lower()
        method_obj = getattr(self, method_name, None)
        if not method_obj:
            raise NotImplementedError(method_name)
        resp = method_obj(*args, **kwargs)

        return resp or self.render()

    def head(self, *args, **kwargs):
        valid_methods = []
        for method in ('get', 'post'):
            if hasattr(self, method):
                valid_methods.append(method.upper())
        return flask.abort(405, valid_methods=valid_methods)


class UserResponderMixin(object):
    flash_invalid_user = _('No user account matches: {}'), 'error'
    flash_unverified_user = _(
        'The user account "{}" has an unverified email address.  Please check'
        ' your email for a verification link from this website.  Or, use the "forgot'
        ' password" link to verify the account.'
    ), 'error'
    flash_disabled_user = _(
        'The user account "{}" has been disabled.  Please contact this'
        ' site\'s administrators for more information.'
    ), 'error'

    def on_inactive_user(self, user):
        if flask.current_app.auth_manager.mail_manager and not user.is_verified:
            if self.flash_unverified_user:
                message, category = self.flash_unverified_user
                flash(message.format(user.email), category)
        if not user.is_enabled:
            self.on_disabled_user(user)

    def on_invalid_user(self, username):
        if self.flash_invalid_user:
            message, category = self.flash_invalid_user
            flash(message.format(username), category)

    def on_disabled_user(self, user):
        if self.flash_disabled_user:
            message, category = self.flash_disabled_user
            flash(message.format(user.display_value), category)


class LoginResponderMixin(UserResponderMixin):
    """ Wrap user authentication view-layer logic

        Flash messages, what to do when a user has been authenticated (by whatever method the
        parent authenticator uses), redirects to a safe URL after login, etc.
    """
    url = '/login'
    flash_success = _('Login successful.'), 'success'

    @staticmethod
    def is_safe_url(target):
        """Returns `True` if the target is a valid URL for redirect"""
        # from http://flask.pocoo.org/snippets/62/
        ref_url = urlparse(flask.request.host_url)
        test_url = urlparse(urljoin(flask.request.host_url, target))
        return (
            test_url.scheme in ('http', 'https')
            and ref_url.netloc == test_url.netloc
        )

    def on_success(self, user):
        flask_login.login_user(user)
        if self.flash_success:
            flash(*self.flash_success)

        # support Flask-Login "next" parameter
        next_parameter = flask.request.values.get('next')
        if flask.current_app.config.get('USE_SESSION_FOR_NEXT'):
            next_parameter = flask.session.get('next')
        if next_parameter and self.is_safe_url(next_parameter):
            redirect_to = next_parameter
        else:
            redirect_to = flask.current_app.auth_manager.url_for('after-login')

        return flask.redirect(redirect_to)


class FormResponderMixin(object):
    """ Wrap form usage for auth responders, contains GET and POST handlers"""
    flash_form_error = _('The form has errors, please see below.'), 'error'
    form_cls = None
    page_title = None

    def on_form_error(self, form):
        if self.flash_form_error:
            flash(*self.flash_form_error)

    def on_form_valid(self, form):
        raise NotImplementedError  # pragma: no cover

    def create_form(self):
        return self.form_cls()

    def assign_template_vars(self, form):
        self.assign('form', form)
        self.assign('form_action_text', self.page_title)
        self.assign('page_title', self.page_title)
        self.assign('page_heading', self.page_title)

    def get(self, *args, **kwargs):
        form = self.create_form()
        self.assign_template_vars(form)

    def post(self, *args, **kwargs):
        form = self.create_form()
        if form.validate():
            resp = self.on_form_valid(form)
            if resp is not None:
                return resp
        else:
            self.on_form_error(form)

        self.assign_template_vars(form)


class AttemptLimitMixin(object):
    @property
    def attempt_ent(self):
        return flask.current_app.auth_manager.entity_registry.attempt_cls

    def should_limit_attempts(self):
        try:
            return self.attempt_ent is not None
        except RegistryError:
            return False

    @property
    def should_filter_ip(self):
        return flask.current_app.config.get('KEGAUTH_ATTEMPT_IP_LIMIT', False)

    def get_input_filters(self, username):
        input_filters = self.attempt_ent.user_input == username
        if self.should_filter_ip and self.get_request_remote_addr():
            input_filters = sa.sql.or_(
                input_filters,
                self.attempt_ent.source_ip == self.get_request_remote_addr()
            )
        return input_filters

    def get_last_limiting_attempt(self, username):
        '''
        Get the last attempt that counts toward the limit count. Attempts that count
        toward the limit before this attempt will be counted to determine if this
        attempt caused a lockout.

        For login, this will be the last failed attempt.
        For password reset, this will be the last attempt.
        '''
        raise NotImplementedError()  # pragma: no cover

    def get_limiting_attempt_count(self, before_time, username):
        '''
        Return the number of attempts that count toward the limit up to before_time.
        '''
        raise NotImplementedError  # pragma: no cover

    def is_attempt_blocked(self, username):
        last_limiting_attempt = self.get_last_limiting_attempt(username)

        if last_limiting_attempt:
            limiting_attempt_count = self.get_limiting_attempt_count(
                last_limiting_attempt.datetime_utc,
                username
            )
            # A failed attempt has caused a lockout so we should check if the lockout
            # period has passed.
            if limiting_attempt_count >= self.get_attempt_limit():
                lockout_delta = timedelta(seconds=self.get_attempt_lockout_period())
                return arrow.utcnow() - last_limiting_attempt.datetime_utc <= lockout_delta

        # We are not in a lockout period so block attempts if there are too many
        # limiting attempts in the timespan until now.
        limiting_in_timespan_count = self.get_limiting_attempt_count(arrow.utcnow(), username)
        return limiting_in_timespan_count >= self.get_attempt_limit()

    def log_attempt(self, username, *, success=True, is_during_lockout=False):
        attempt = self.attempt_ent(
            attempt_type=self.get_attempt_type(),
            user_input=username,
            success=success,
            is_during_lockout=is_during_lockout,
            datetime_utc=arrow.utcnow(),
        )
        if flask.has_request_context():
            attempt.source_ip = self.get_request_remote_addr()

        db.session.add(attempt)
        db.session.commit()
        return attempt

    def update_attempt(self, attempt, **kwargs):
        self.attempt_ent.edit(attempt.id, **kwargs)

    @staticmethod
    def get_request_remote_addr():
        return flask.request.remote_addr

    def on_attempt_blocked(self):
        flash_message = self.get_flash_attempts_limit_reached()
        if flash_message:
            flash(*flash_message)

    def get_flash_attempts_limit_reached(self):
        raise NotImplementedError  # pragma: no cover

    def get_attempt_limit(self):
        raise NotImplementedError  # pragma: no cover

    def get_attempt_timespan(self):
        raise NotImplementedError  # pragma: no cover

    def get_attempt_type(self):
        raise NotImplementedError  # pragma: no cover

    def get_attempt_lockout_period(self):
        raise NotImplementedError  # pragma: no cover


class PasswordSetterResponderBase(FormResponderMixin, ViewResponder):
    """ Base logic for resetting passwords and verifying accounts via token"""
    form_cls = forms.SetPassword
    template_name = 'keg_auth/set-password.html'
    flash_invalid_token = _(
        'Authentication token was invalid or expired.  Please fill out the'
        ' form below to get a new token.'
    ), 'error'

    def __call__(self, *args, **kwargs):
        if not flask.current_app.auth_manager.mail_manager:
            flask.abort(404)

        self.user_loader(kwargs.get('user_id'))
        self.token = kwargs.get('token')
        self.pre_method()

        return super(PasswordSetterResponderBase, self).__call__(*args, **kwargs)

    def flash_and_redirect(self, flash_parts, auth_ident):
        if flash_parts:
            flash(*flash_parts)
        redirect_to = flask.current_app.auth_manager.url_for(auth_ident)
        flask.abort(flask.redirect(redirect_to))

    def user_loader(self, user_id):
        user_ent = flask.current_app.auth_manager.entity_registry.user_cls
        self.user = user_ent.query.get(user_id)
        if not self.user:
            flask.abort(404)

    def create_form(self):
        return self.form_cls(user=self.user)

    def pre_method(self):
        if not self.user.token_verify(self.token):
            resp = self.on_invalid_token()
            # In case on_invalid_token() is replaced and it accidently fails to return a value
            # make sure we change that to a generic 400.
            flask.abort(resp or 400)

    def on_form_valid(self, form):
        new_password = form.password.data
        self.user.change_password(self.token, new_password)
        self.flash_and_redirect(self.flash_success, self.on_success_endpoint)

    def on_invalid_token(self):
        self.flash_and_redirect(self.flash_invalid_token, 'forgot-password')

    def assign_template_vars(self, form):
        super(PasswordSetterResponderBase, self).assign_template_vars(form)
        self.assign('submit_button_text', self.submit_button_text)


class ResetPasswordViewResponder(AttemptLimitMixin, PasswordSetterResponderBase):
    """ Responder for resetting passwords via token on keg-auth logins"""
    url = '/reset-password/<int:user_id>/<token>'
    page_title = _('Complete Password Reset')
    submit_button_text = _('Change Password')
    flash_success = _('Password changed.  Please use the new password to login below.'), 'success'
    on_success_endpoint = 'after-reset'

    def on_form_valid(self, form):
        if self.should_limit_attempts():
            if self.is_attempt_blocked(get_username(self.user)):
                self.log_attempt(get_username(self.user), success=False, is_during_lockout=True)
                self.on_attempt_blocked()
                return

            self.log_attempt(get_username(self.user), success=True)

        new_password = form.password.data
        self.user.change_password(self.token, new_password)
        self.flash_and_redirect(self.flash_success, self.on_success_endpoint)

    def get_flash_attempts_limit_reached(self):
        return _('Too many password reset attempts.'), 'error'

    def get_attempt_limit(self):
        return flask.current_app.config.get('KEGAUTH_RESET_ATTEMPT_LIMIT')

    def get_attempt_timespan(self):
        return flask.current_app.config.get('KEGAUTH_RESET_ATTEMPT_TIMESPAN')

    def get_attempt_lockout_period(self):
        return flask.current_app.config.get('KEGAUTH_RESET_ATTEMPT_LOCKOUT')

    def get_attempt_type(self):
        return 'reset'

    def get_last_limiting_attempt(self, username):
        return self.attempt_ent.query.filter_by(
            is_during_lockout=False,
            attempt_type=self.get_attempt_type(),
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

    def get_limiting_attempt_count(self, before_time, username):
        timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())
        return self.attempt_ent.query.filter(
            self.get_input_filters(username),
            self.attempt_ent.is_during_lockout == sa.false(),
            self.attempt_ent.datetime_utc > timespan_start,
            self.attempt_ent.datetime_utc <= before_time,
            self.attempt_ent.attempt_type == self.get_attempt_type(),
        ).count()


class VerifyAccountViewResponder(PasswordSetterResponderBase):
    """ Responder for verifying users via email token for keg-auth logins"""
    url = '/verify-account/<int:user_id>/<token>'
    page_title = _('Verify Account & Set Password')
    submit_button_text = _('Verify & Set Password')
    flash_success = _('Account verified & password set.  Please use the new password to login'
                      ' below.'), 'success'
    on_success_endpoint = 'after-verify-account'


class PasswordFormViewResponder(AttemptLimitMixin, LoginResponderMixin,
                                FormResponderMixin, ViewResponder):
    """ Master responder for username/password-style logins, using a login form"""
    template_name = 'keg_auth/login.html'
    page_title = _('Log In')
    flash_invalid_password = _('Invalid password.'), 'error'

    @property
    def form_cls(self):
        return forms.login_form()

    def on_form_valid(self, form):
        username = form.login_id.data
        attempt = None
        if self.should_limit_attempts():
            if self.is_attempt_blocked(username):
                # If we are rate-limiting this attempt, we don't want to proceed with validation.
                # Validating may still allow brute forcing by measuring response time. Skipping it
                # may help mitigate DoS attacks on the login page as password hashing is typically
                # an expensive operation
                attempt = self.log_attempt(username, success=False, is_during_lockout=True)
                self.on_attempt_blocked()
                return
            else:
                attempt = self.log_attempt(username, success=False)

        try:
            # We want to know if the login attempt was successful so we'll try
            # to verify the user. If the user is verified but the attempt is blocked,
            # mark the attempt as successful and abort.
            user = self.parent.verify_user(
                login_id=form.login_id.data,
                password=form.password.data
            )

            if attempt:
                self.update_attempt(attempt, success=True)

            # User is active and password is verified
            return self.on_success(user)
        except UserNotFound:
            self.on_invalid_user(form.login_id.data)
        except UserInactive as exc:
            self.on_inactive_user(exc.user)
        except UserInvalidAuth as exc:
            self.on_invalid_password(exc.user)

    def on_invalid_password(self, user):
        if self.flash_invalid_password:
            flash(*self.flash_invalid_password)

    def get_flash_attempts_limit_reached(self):
        return _('Too many failed login attempts.'), 'error'

    def get_attempt_limit(self):
        return flask.current_app.config.get('KEGAUTH_LOGIN_ATTEMPT_LIMIT')

    def get_attempt_timespan(self):
        return flask.current_app.config.get('KEGAUTH_LOGIN_ATTEMPT_TIMESPAN')

    def get_attempt_lockout_period(self):
        return flask.current_app.config.get('KEGAUTH_LOGIN_ATTEMPT_LOCKOUT')

    def get_attempt_type(self):
        return 'login'

    def get_last_limiting_attempt(self, username):
        return self.attempt_ent.query.filter_by(
            success=False,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type(),
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

    def get_limiting_attempt_count(self, before_time, username):
        last_successful_attempt = self.attempt_ent.query.filter_by(
            success=True,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type()
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

        def is_within_timespan(attempt):
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())
            return attempt.datetime_utc > timespan_start

        if last_successful_attempt and is_within_timespan(last_successful_attempt):
            timespan_start = last_successful_attempt.datetime_utc
        else:
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())

        return self.attempt_ent.query.filter(
            self.get_input_filters(username),
            self.attempt_ent.success == sa.false(),
            self.attempt_ent.is_during_lockout == sa.false(),
            self.attempt_ent.datetime_utc > timespan_start,
            self.attempt_ent.datetime_utc <= before_time,
            self.attempt_ent.attempt_type == self.get_attempt_type(),
        ).count()


class ForgotPasswordViewResponder(AttemptLimitMixin, UserResponderMixin, FormResponderMixin,
                                  ViewResponder):
    """ Master responder for keg-integrated logins, using an email form"""
    url = '/forgot-password'
    form_cls = forms.ForgotPassword
    page_title = _('Initiate Password Reset')
    template_name = 'keg_auth/forgot-password.html'
    flash_success = _('Please check your email for the link to change your password.'), 'success'

    def __call__(self, *args, **kwargs):
        if not flask.current_app.auth_manager.mail_manager:
            flask.abort(404)

        return super(ForgotPasswordViewResponder, self).__call__(*args, **kwargs)

    def on_form_valid(self, form):
        attempt = None
        if self.should_limit_attempts():
            if self.is_attempt_blocked(form.email.data):
                # If we are rate-limiting this attempt, we don't want to proceed with validation.
                # Validating may still allow brute forcing by measuring response time.
                attempt = self.log_attempt(form.email.data, success=False, is_during_lockout=True)
                self.on_attempt_blocked()
                return
            else:
                attempt = self.log_attempt(form.email.data, success=False)

        try:
            user = self.parent.verify_user(login_id=form.email.data, allow_unverified=True)

            if attempt:
                self.update_attempt(attempt, success=True)

            # User is active, take action to initiate password reset
            return self.on_success(user)
        except UserNotFound:
            self.on_invalid_user(form.email.data)
        except UserInactive as exc:
            self.on_disabled_user(exc.user)

    def on_success(self, user):
        self.send_email(user)
        if self.flash_success:
            flash(*self.flash_success)
        redirect_to = flask.current_app.auth_manager.url_for('after-forgot')
        return flask.redirect(redirect_to)

    def send_email(self, user):
        user.token_generate()
        flask.current_app.auth_manager.mail_manager.send_reset_password(user)

    def get_flash_attempts_limit_reached(self):
        return _('Too many failed attempts.'), 'error'

    def get_attempt_limit(self):
        return flask.current_app.config.get('KEGAUTH_FORGOT_ATTEMPT_LIMIT')

    def get_attempt_timespan(self):
        return flask.current_app.config.get('KEGAUTH_FORGOT_ATTEMPT_TIMESPAN')

    def get_attempt_lockout_period(self):
        return flask.current_app.config.get('KEGAUTH_FORGOT_ATTEMPT_LOCKOUT')

    def get_attempt_type(self):
        return 'forgot'

    def get_last_limiting_attempt(self, username):
        return self.attempt_ent.query.filter_by(
            success=False,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type(),
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

    def get_limiting_attempt_count(self, before_time, username):
        last_successful_attempt = self.attempt_ent.query.filter_by(
            success=True,
            is_during_lockout=False,
            attempt_type=self.get_attempt_type()
        ).filter(
            self.get_input_filters(username)
        ).order_by(
            self.attempt_ent.datetime_utc.desc(),
        ).first()

        def is_within_timespan(attempt):
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())
            return attempt.datetime_utc > timespan_start

        if last_successful_attempt and is_within_timespan(last_successful_attempt):
            timespan_start = last_successful_attempt.datetime_utc
        else:
            timespan_start = before_time + timedelta(seconds=-self.get_attempt_timespan())

        return self.attempt_ent.query.filter(
            self.get_input_filters(username),
            self.attempt_ent.success == sa.false(),
            self.attempt_ent.is_during_lockout == sa.false(),
            self.attempt_ent.datetime_utc > timespan_start,
            self.attempt_ent.datetime_utc <= before_time,
            self.attempt_ent.attempt_type == self.get_attempt_type(),
        ).count()


class LogoutViewResponder(ViewResponder):
    url = '/logout'
    flash_success = _('You have been logged out.'), 'success'

    def get(self):
        flask_login.logout_user()
        if self.flash_success:
            flash(*self.flash_success)
        redirect_to = flask.current_app.auth_manager.url_for('after-logout')
        flask.abort(flask.redirect(redirect_to))


class PasswordAuthenticatorMixin(object):
    """ Username/password authenticators will need a way to verify a user is valid
        prior to making it the current user in flask login """
    responder_cls = {
        'login': PasswordFormViewResponder
    }

    def verify_user(self, login_id=None, password=None):
        raise NotImplementedError

    def verify_password(self, user, password):
        return NotImplementedError


class TokenLoaderMixin(object):
    """ Token authenticators will need a way to generate an access token, which will then be
        loaded in the request to log a user into flask-login """
    authentication_failure_redirect = False

    def create_access_token(self, user):
        raise NotImplementedError


class KegAuthenticator(PasswordAuthenticatorMixin, LoginAuthenticator):
    """ Uses username/password authentication with a login form, validates against keg-auth db"""
    responder_cls = {
        'login': PasswordFormViewResponder,
        'forgot-password': ForgotPasswordViewResponder,
        'reset-password': ResetPasswordViewResponder,
        'verify-account': VerifyAccountViewResponder,
        'logout': LogoutViewResponder,
    }

    def verify_user(self, login_id=None, password=None, allow_unverified=False):
        user = self.user_ent.query.filter_by(username=login_id).one_or_none()

        if not user:
            raise UserNotFound
        if not allow_unverified and not user.is_active:
            raise UserInactive(user)
        if allow_unverified and not user.is_enabled:
            raise UserInactive(user)
        if password and not self.verify_password(user, password):
            raise UserInvalidAuth(user)

        return user

    def verify_password(self, user, password):
        return user.password == password


class LdapAuthenticator(KegAuthenticator):
    """ Uses username/password authentication with a login form, validates against LDAP host

        Most responder types won't be relevant here.
    """
    def verify_user(self, login_id=None, password=None):
        user = self.user_ent.query.filter_by(username=login_id).one_or_none()

        if not user:
            username_key = get_username_key(self.user_ent)
            user = self.user_ent.add(**{username_key: login_id})
        if password and not self.verify_password(user, password):
            raise UserInvalidAuth(user)

        return user

    def verify_password(self, user, password):
        """
        Check the given username/password combination at the application's configured LDAP server.
        Returns `True` if the user authentication is successful, `False` otherwise. NOTE: By
        request, authentication can be bypassed by setting the KEGAUTH_LDAP_TEST_MODE configuration
        setting to `True`. When set, all authentication attempts will succeed!
        """

        if flask.current_app.config.get('KEGAUTH_LDAP_TEST_MODE', False):
            return True

        ldap_url = flask.current_app.config.get('KEGAUTH_LDAP_SERVER_URL')
        if not ldap_url:
            raise Exception(_('No KEGAUTH_LDAP_SERVER_URL configured!'))

        ldap_dn_format = flask.current_app.config.get('KEGAUTH_LDAP_DN_FORMAT')
        if not ldap_dn_format:
            raise Exception(_('No KEGAUTH_LDAP_DN_FORMAT configured!'))

        def ldap_bind(server_url):
            session = ldap.initialize(server_url)

            try:
                dn = ldap_dn_format.format(user.username)
                result = session.simple_bind_s(dn, password)
                del session
                return bool(
                    result
                    and len(result)
                    and result[0] == ldap.RES_BIND
                )
            except (ldap.INVALID_CREDENTIALS, ldap.INVALID_DN_SYNTAX):
                return False

        if isinstance(ldap_url, str):
            return ldap_bind(ldap_url)

        # We have a list of servers.
        for server_url in ldap_url:
            if ldap_bind(server_url):
                return True

        return False


class OidcLoginViewResponder(LoginResponderMixin, ViewResponder):
    """ OIDC logins, using an oauth token"""

    flash_success = None
    page_title = 'Log In'
    template_name = 'keg_auth/flash-messages-only.html'

    def get(self, *args, **kwargs):
        oidc = flask.current_app.auth_manager.oidc
        oidc_check = oidc.require_login(lambda: True)()
        if oidc_check is not True:
            return oidc_check

        login_id = oidc.user_getfield("preferred_username")

        try:
            user = self.parent.verify_user(login_id=login_id)

            # User is active and password is verified
            return self.on_success(user)
        except UserNotFound:
            self.on_invalid_user(login_id)
        except UserInactive as exc:
            self.on_inactive_user(exc.user)

    def head(self, *args, **kwargs):
        return flask.abort(405, valid_methods=['GET'])

    def post(self, *args, **kwargs):
        return flask.abort(405, valid_methods=['GET'])


class OidcLogoutViewResponder(LogoutViewResponder):
    """ OIDC logout requires some extra leg-work, because token gets refreshed server-side"""

    def get(self):
        oidc = flask.current_app.auth_manager.oidc
        url_login = flask.url_for(flask.current_app.auth_manager.endpoint('login'))
        url_after_login = flask.url_for(flask.current_app.auth_manager.endpoint('after-login'))
        bad_token_redirect_resp = flask.current_app.login_manager.unauthorized()

        url_only_redirect = flask.current_app.config.get('KEGAUTH_OIDC_LOGOUT_REDIRECT')
        if url_only_redirect:
            flask_login.logout_user()
            return flask.abort(flask.redirect(url_only_redirect))

        """ Logout won't work if user isn't authenticated to begin with, i.e. there won't be a
            token to use. Just redirect to a sane place to force a login to continue."""
        try:
            user_sub = oidc.user_getfield('sub')
        except Exception as exc:
            if 'User was not authenticated' not in str(exc):
                raise
            return flask.abort(flask.redirect(url_login))

        """ In some cases e.g. app restart, credentials store may not have valid information in the
            flask server-side info. In that case, clear the client token and refresh info from
            the oauth source. We have to have a valid id token to make logout work."""
        try:
            from oauth2client.client import OAuth2Credentials
            id_token = OAuth2Credentials.from_json(
                oidc.credentials_store[user_sub]
            ).token_response['id_token']
        except KeyError:
            oidc.logout()
            return flask.abort(bad_token_redirect_resp)

        """ Build the oauth request URI, which has to include the ID token. But, logout all client
            session info before redirecting there."""
        logout_request = '{}{}?id_token_hint={}&post_logout_redirect_uri={}'.format(
            flask.current_app.config.get('OIDC_PROVIDER_URL'),
            flask.current_app.config.get('OIDC_LOGOUT'),
            str(id_token),
            flask.current_app.config.get('OIDC_REDIRECT_BASE') + url_after_login,
        )
        oidc.logout()
        flask_login.logout_user()
        return flask.redirect(logout_request)


class OidcAuthenticator(LoginAuthenticator):
    """ Uses OIDC authentication with an oauth provider, validates against keg-auth db"""
    responder_cls = {
        'login': OidcLoginViewResponder,
        'logout': OidcLogoutViewResponder,
    }

    def __init__(self, app):
        from flask_oidc import OpenIDConnect

        oidc_settings = {
            'web': {
                'client_id': app.config.get('OIDC_CLIENT_ID'),
                'client_secret': app.config.get('OIDC_CLIENT_SECRET'),
                'auth_uri': app.config.get('OIDC_PROVIDER_URL') + app.config.get('OIDC_AUTH_URI'),
                'token_uri': app.config.get('OIDC_PROVIDER_URL') + app.config.get('OIDC_TOKEN_URI'),
                'issuer': app.config.get('OIDC_PROVIDER_URL') + app.config.get('OIDC_ISSUER'),
                'userinfo_uri': app.config.get('OIDC_PROVIDER_URL')
                + app.config.get('OIDC_USERINFO_URI'),
                'redirect_uris': [
                    app.config.get('OIDC_REDIRECT_BASE') + app.config.get('OIDC_CALLBACK_ROUTE')
                ]
            }
        }

        class KAOpenIDConnect(OpenIDConnect):
            def load_secrets(self, app):
                return oidc_settings
        app.auth_manager.oidc = KAOpenIDConnect(app)

        super().__init__(app)

    def verify_user(self, login_id=None):
        user = self.user_ent.query.filter_by(username=login_id).one_or_none()

        if not user:
            raise UserNotFound
        if not user.is_active:
            raise UserInactive(user)

        return user


class JwtRequestLoader(TokenLoaderMixin, RequestLoader):
    """ Loader for JWT tokens contained in the Authorization header.

        Requires flask-jwt-extended (`pip install keg-auth[jwt]`)"""
    def __init__(self, app):
        super(JwtRequestLoader, self).__init__(app)

        self.jwt_manager = jwt_manager = flask_jwt_extended.JWTManager()
        jwt_manager.init_app(app)

        @jwt_manager.user_identity_loader
        def user_identity_loader(user):
            """
            Serialize a user entity to the JWT token
            This method is the complement of `user_lookup_loader`
            """
            return user.session_key

        @jwt_manager.user_lookup_loader
        def user_loader_callback_loader(jwt_header, jwt_data):
            """
            Load a user entity from the JWT token
            This method is the complement of `user_identity_loader`

            Note, if user is not found or inactive, fail silently - user just won't get loaded
            """
            data_key = flask.current_app.config.get('JWT_IDENTITY_CLAIM')
            return self.user_ent.get_by(session_key=jwt_data[data_key], is_active=True)

    @staticmethod
    def get_authenticated_user():
        try:
            if flask_jwt_extended.verify_jwt_in_request() is None:
                return None
            user = flask_jwt_extended.get_current_user()
            flask_login.login_user(user)
            return user
        except flask_jwt_extended.exceptions.JWTExtendedException:
            return None

    def create_access_token(self, user):
        return flask_jwt_extended.create_access_token(user)


class TokenRequestLoader(RequestLoader):
    authentication_failure_redirect = False

    def get_authenticated_user(self):
        token = flask.request.headers.get('X-Auth-Token')

        if token is None:
            return

        user = self.user_ent.get_by_token(token)

        if user is None:
            return

        flask_login.login_user(user)
        return user


class PasswordCharset(typing.NamedTuple):
    name: str
    alphabet: str


class PasswordPolicyError(Exception):
    pass


PasswordCheckFunction = typing.Callable[[str, typing.Any], None]


class PasswordPolicy:
    """
    A base class that defines password requirements for the application.
    This class defines some basic, common validations and can be extended or limited by subclassing.

    To define additional password checks, create a method on your subclass that accepts a
    password string and a user entity object and raises PasswordPolicyError if the password does
    not meet the requirement you intend to check. Then override password_checks to add your method
    to the returned list of methods.

    To remove a password check that is enabled by default, override password_checks and return only
    the methods you wish to use.

    Default settings are based on NIST guidelines and some common restrictions.
    """

    """
    Minimum password length for check_length validation
    """
    min_length: int = 8

    """
    Character sets used for checking minimum "complexity" in check_character_set validation
    """
    required_char_types: typing.List[PasswordCharset] = [
        PasswordCharset(_('lowercase letter'), string.ascii_lowercase),
        PasswordCharset(_('uppercase letter'), string.ascii_uppercase),
        PasswordCharset(_('number'), string.digits),
        PasswordCharset(_('symbol'), ''.join(sorted(
            set(string.printable)
            - set(string.whitespace + string.ascii_letters + string.digits))
        )),
    ]

    """
    Minimum character number of different character types required in check_character_set validation
    """
    required_min_char_types: int = 3

    def check_length(self, pw: str, user):
        """
        Raises PasswordPolicyError if a password is not at least min_length characters long.
        :param pw: password to check
        :param user: user entity
        """
        if len(pw) < self.min_length:
            raise PasswordPolicyError(_(
                'Password must be at least {min_length} characters long',
                min_length=self.min_length
            ))

    def check_character_set(self, pw: str, user):
        """
        Raises PasswordPolicyError if a password does not contain at least one character from
        at least `required_at_least_char_types` of the alphabets in `required_char_sets`.
        :param pw: password to check
        :param user: user entity
        """

        missing = []
        for name, alphabet in self.required_char_types:
            if not set(pw) & set(alphabet):
                missing.append(name)

        if len(missing) > len(self.required_char_types) - self.required_min_char_types:
            if len(self.required_char_types) == 1:
                message = _('Password must include a {type}', type=self.required_char_types[0].name)
            else:
                first_part = ', '.join(str(t.name) for t in self.required_char_types[:-1])
                message = _(
                    'Password must include at least {required} of {first} and/or {last}',
                    required=self.required_min_char_types,
                    first=first_part,
                    last=self.required_char_types[-1].name
                )

            raise PasswordPolicyError(message)

    def check_does_not_contain_username(self, pw: str, user):
        """
        Raises PasswordPolicyError if the password contains the username. This is case insensitive.
        :param pw: password to check
        :param user: user entity
        """
        user_cls = user.__class__
        username_key = get_username_key(user_cls)
        username = getattr(user, username_key)
        username_col = getattr(user_cls, username_key)
        if isinstance(username_col.type, EmailType):
            username = user.username.split('@')[0]

        if username.casefold() in pw.casefold():
            raise PasswordPolicyError(_('Password may not contain username'))

    def password_checks(self) -> typing.List[PasswordCheckFunction]:
        return [
            self.check_length,
            self.check_character_set,
            self.check_does_not_contain_username,
        ]

    @classmethod
    def generate_validator(cls, check: PasswordCheckFunction) -> typing.Callable:
        def validator(form: wtforms.Form, field: wtforms.Field):
            try:
                check(field.data, form.user)
            except PasswordPolicyError as e:
                raise wtforms.ValidationError(str(e))
        return validator

    @classmethod
    def form_validators(cls):
        policy = cls()
        return [cls.generate_validator(c) for c in policy.password_checks()]


class DefaultPasswordPolicy(PasswordPolicy):
    """
    A bare-bones, very permissive policy to use as a default if none is set on initialization.
    """
    def password_checks(self):
        return [self.check_length]



import base64
import binascii
import hashlib
import json

import arrow
import flask
import itsdangerous
import keg_elements.db.utils as dbutils
import passlib.hash
import passlib.pwd
import shortuuid
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm
import sqlalchemy.sql as sa_sql
from blazeutils import tolist
from blazeutils.strings import randchars
from keg.db import db
from keg_elements.db.mixins import might_commit, might_flush
from sqlalchemy.dialects import mssql
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy_utils import (
    ArrowType,
    EmailType,
    PasswordType,
    force_auto_coercion,
)

from keg_auth.model.types import AttemptType

force_auto_coercion()


def registry():
    return flask.current_app.auth_manager.entity_registry


def _create_cryptcontext_kwargs(**column_kwargs):
    config = flask.current_app.config['PASSLIB_CRYPTCONTEXT_KWARGS']
    retval = {}
    retval.update(config)
    retval.update(column_kwargs)
    return retval


def _generate_session_key():
    return str(shortuuid.uuid())


class InvalidToken(Exception):
    pass


class KAPasswordType(PasswordType):
    def load_dialect_impl(self, dialect):
        if dialect.name == 'mssql':
            return mssql.VARCHAR(self.length)
        return super(KAPasswordType, self).load_dialect_impl(dialect)


class UserMixin(object):
    """Generic mixin for user entities."""
    # These two attributes are needed by Flask-Login.
    is_anonymous = False
    is_authenticated = True

    is_enabled = sa.Column(sa.Boolean, nullable=False, default=True)
    is_superuser = sa.Column(sa.Boolean, nullable=False, default=False)
    password = sa.Column(KAPasswordType(onload=_create_cryptcontext_kwargs))

    username = sa.Column(sa.Unicode(512), nullable=False, unique=True)

    # key used to identify the "id" for flask-login, which is expected to be a string. While we
    #   could return the user's db id cast to string, that would not give us a hook to reset
    #   sessions when permissions go stale
    session_key = sa.Column(sa.Unicode(36), nullable=False, unique=True,
                            default=_generate_session_key)

    # When a user logins we need to track their last login time
    # This is used in the salt to invalidate a password/verification token
    # when a user logs in.
    last_login_utc = sa.Column(ArrowType, nullable=True, default=None, server_default=None)

    # The datetime when a user will be disabled. User will be inactive if this is set
    # to a datetime in the past.
    disabled_utc = sa.Column(ArrowType, nullable=True, default=None, server_default=sa.null())

    # is_active defines the complexities of how to determine what users are active. For instance,
    #   if email is in scope, we need to have an additional flag to verify users, and that would
    #   get included in is_active logic.
    @hybrid_property
    def is_active(self):
        return not self.is_disabled_by_date and self.is_enabled

    @is_active.expression
    def is_active(cls):
        expr = sa_sql.and_(~cls.is_disabled_by_date, cls.is_enabled == sa.true())
        # need to wrap the expression in a case to work with MSSQL
        return sa_sql.case([(expr, sa.true())], else_=sa.false())

    @hybrid_property
    def is_disabled_by_date(self):
        return self.disabled_utc is not None and self.disabled_utc <= arrow.utcnow()

    @is_disabled_by_date.expression
    def is_disabled_by_date(cls):
        is_disabled_expr = sa.sql.and_(
            cls.disabled_utc.isnot(None),
            cls.disabled_utc <= arrow.utcnow(),
        )
        return sa_sql.case([(is_disabled_expr, sa.true())], else_=sa.false())

    def get_id(self):
        # Flask-Login requires that this return a string value for the session
        return str(self.session_key)

    def reset_session_key(self):
        self.session_key = _generate_session_key()

    @property
    def display_value(self):
        # shortcut to return the value of the user ident attribute
        return self.username

    @classmethod
    def testing_create(cls, **kwargs):
        kwargs['password'] = kwargs.get('password') or randchars()

        if 'permissions' in kwargs:
            perm_cls = registry().permission_cls

            # ensure all of the tokens exists
            flask.current_app.auth_manager.validate_permission_set(
                list(filter(
                    lambda perm: not isinstance(perm, perm_cls),
                    tolist(kwargs['permissions'])
                ))
            )

            kwargs['permissions'] = [
                perm_cls.testing_create(token=perm)
                if not isinstance(perm, perm_cls) else perm
                for perm in tolist(kwargs['permissions'])
            ]

        user = super(UserMixin, cls).testing_create(**kwargs)
        user._plaintext_pass = kwargs['password']
        return user

    def get_all_permissions(self):
        # Superusers are considered to have all permissions.
        if self.is_superuser:
            return set(registry().permission_cls.query)

        perm_cls = registry().permission_cls
        mapping = self._query_permission_mapping().alias('user_permission_mapping')
        q = db.session.query(
            perm_cls
        ).select_from(
            perm_cls
        ).join(
            mapping,
            mapping.c.perm_id == perm_cls.id
        ).filter(
            mapping.c.user_id == self.id
        )
        return set(q)

    def get_all_permission_tokens(self):
        # permission tokens for a given user should get loaded once per session. This method, called
        #   by has_all_permissions, is the main interface to grab them. So, set up a cache here, so
        #   the user instance stored by flask-login will hold the set to be used (rather than
        #   continuing to query the database on each permission check)
        # the other side to this is that permissions can become stale, because we are not querying
        #   the database every time. If an admin changes permissions while a user is actively
        #   logged in, we have to make sure the session is invalidated (see session_key field)
        if not hasattr(self, '_permission_cache'):
            self._permission_cache = {p.token for p in self.get_all_permissions()}
        return self._permission_cache

    def has_all_permissions(self, *tokens):
        return set(tokens).issubset(self.get_all_permission_tokens())

    def has_any_permission(self, *tokens):
        return bool(set(tokens).intersection(self.get_all_permission_tokens()))

    @classmethod
    def _query_permission_mapping(cls):
        return sa.union(
            cls._query_direct_permissions(),
            cls._query_bundle_permissions(),
            cls._query_group_permissions()
        )

    @classmethod
    def _query_direct_permissions(cls):
        perm_cls = registry().permission_cls
        return db.session.query(
            cls.id.label('user_id'),
            perm_cls.id.label('perm_id'),
        ).select_from(
            cls
        ).join(
            cls.permissions
        )

    @classmethod
    def _query_bundle_permissions(cls):
        perm_cls = registry().permission_cls
        bundle_cls = registry().bundle_cls
        return db.session.query(
            cls.id.label('user_id'),
            perm_cls.id.label('perm_id'),
        ).select_from(
            cls
        ).join(
            cls.bundles
        ).join(
            bundle_cls.permissions
        )

    @classmethod
    def _query_group_permissions(cls):
        group_cls = registry().group_cls

        group_mapping = group_cls._query_permission_mapping().alias('group_permissions_mapping')
        return db.session.query(
            cls.id.label('user_id'),
            group_mapping.c.perm_id.label('perm_id')
        ).select_from(
            cls
        ).join(
            cls.groups
        ).join(
            group_mapping,
            group_mapping.c.group_id == group_cls.id
        )

    def get_token_salt(self):
        """
        Create salt data for password reset token signing. The return value will be hashed
        together with the signing key. This ensures that changes to any of the fields included
        in the salt invalidates any tokens produced with the old values
        Values included:

            * user login identifier -> if username/email change it will invalidate
                                       the user token
            * is_active -> Anytime a user verifies will invalidate a token

            * current password hash or empty string if no password has been set
              -> If the password is updated we want to invalidate the token

            * last login time -> Any time a user logs in it will invalidate any
                                 verification and reset password emails

        :return: JSON string of list containing the values listed above
        """
        return json.dumps([
            self.display_value,
            str(self.is_active),
            self.password.hash.decode() if self.password is not None else '',
            self.last_login_utc.to('UTC').isoformat() if self.last_login_utc else None
        ])

    def get_token_serializer(self, expires_in):
        """
        Create a JWT serializer using itsdangerous
        :param expires_in: seconds from token creation until expiration
        :return: TimedJSONWebSignatureSerializer instance
        """
        return itsdangerous.TimedJSONWebSignatureSerializer(
            flask.current_app.config['SECRET_KEY'],
            algorithm_name='HS512',
            expires_in=expires_in,
            signer_kwargs={'digest_method': hashlib.sha512}
        )

    def token_verify(self, token):
        """
        Verify a password reset token. The token is validated for:
            * user identity
            * tampering
            * expiration
            * password was not already reset since token was generated
            * user has not signed in since token was generated

        :param token: string representation of token to verify
        :return: bool indicating token validity
        """
        if not token:
            return False
        if isinstance(token, str):
            token = token.encode()

        serializer = self.get_token_serializer(None)
        try:
            payload = serializer.loads(token, salt=self.get_token_salt())
        except itsdangerous.BadSignature:
            return False
        return payload['user_id'] == self.id

    def token_generate(self):
        """
        Create a new token for this user. The returned value is an expiring JWT
        signed with the application's crypto key. Externally this token should be treated as opaque.
        The value returned by this function must not be persisted.
        :return: a string representation of the generated token
        """
        serializer = self.get_token_serializer(
            flask.current_app.config['KEGAUTH_TOKEN_EXPIRE_MINS'] * 60
        )
        payload = {
            'user_id': self.id,
        }
        token = serializer.dumps(payload, salt=self.get_token_salt()).decode()

        # Store the plain text version on this instance for ease of use.  It will not get
        # pesisted to the db, so no security conern.
        self._token_plain = token

        return token


class UserTokenMixin(object):
    """Mixin for users who will be authenticated by tokens."""
    token = sa.Column(KAPasswordType(onload=_create_cryptcontext_kwargs))

    @classmethod
    def generate_raw_auth_token(cls, length=32, entropy=None, charset='ascii_50'):
        """Return a raw authentication token

        NOTE(nZac): You should not store this directly in the database. When using this mixin,
        simply setting this value to ``self.token = generate_raw_auth_token`` is enough (though,
        there is a helper method for that ``reset_auth_token``).
        """
        return passlib.pwd.genword(length=length, entropy=entropy, charset=charset)

    @classmethod
    def get_user_for_api_token(cls, api_token):
        if api_token is None:
            return

        if isinstance(api_token, bytes):
            api_token = api_token.decode()

        if len(api_token.split('.')) != 2:
            return

        raw_email, raw_token = api_token.split('.')
        try:
            real_email = base64.urlsafe_b64decode(raw_email.encode()).decode()
        except (binascii.Error, TypeError):
            return

        user = cls.query.filter_by(email=real_email).one_or_none()
        if user is None or not user.token.context.verify(raw_token, user.token.hash):
            return
        else:
            return user

    def reset_auth_token(self, **kwargs):
        """Reset the authentication token for this user

        Takes the same parameter as `:cls:generate_auth_token`
        """
        self.token = raw = self.generate_raw_auth_token(**kwargs)
        return raw

    def verify_token(self, token):
        if not token or not self.token:
            return False

        return self.token.context.verify(token, self.token.hash)

    def generate_api_token(self, token=None):
        raw_token = token or self.reset_auth_token()

        url_safe_email = base64.urlsafe_b64encode(self.email.encode()).decode()

        raw_api_token = '{email}.{token}'.format(
            email=url_safe_email,
            token=raw_token,
        )

        return raw_api_token


class UserEmailMixin(object):
    """Mixin for users who will be authenticated by email/password."""
    # Assume the user will need to verify their email address before they become active.
    is_verified = sa.Column(sa.Boolean, nullable=False, default=False)
    email = sa.Column(EmailType, nullable=False, unique=True)

    @hybrid_property
    def username(self):
        return self.email

    @username.expression
    def username(cls):
        return cls.email

    @hybrid_property
    def is_active(self):
        return not self.is_disabled_by_date and self.is_verified and self.is_enabled

    @is_active.expression
    def is_active(cls):
        expr = sa_sql.and_(
            ~cls.is_disabled_by_date,
            cls.is_verified == sa.true(),
            cls.is_enabled == sa.true(),
        )
        # need to wrap the expression in a case to work with MSSQL
        return sa_sql.case([(expr, sa.true())], else_=sa.false())

    @classmethod
    def testing_create(cls, **kwargs):
        # Most tests will want an active user by default, which is the opposite of what we want in
        # production, so swap that logic.
        kwargs.setdefault('is_verified', True)

        user = super(UserEmailMixin, cls).testing_create(**kwargs)
        return user

    @might_commit
    def change_password(self, token, new_password):
        """
            Change a password based on token authorization.
        """
        if not self.token_verify(token):
            raise InvalidToken

        self.password = new_password
        self.is_verified = True


class PermissionMixin(object):
    """Generic mixin for permissions."""
    token = sa.Column(sa.Unicode(1024), nullable=False, unique=True)
    description = sa.Column(sa.Unicode)

    @classmethod
    def get_by_token(cls, token):
        return cls.get_by(token=token)

    @classmethod
    def testing_create(cls, **kwargs):
        matching = None
        if 'token' in kwargs:
            matching = cls.get_by_token(kwargs['token'])
        return matching or super(PermissionMixin, cls).testing_create(**kwargs)


class BundleMixin(object):
    """Generic mixin for permission bundles."""
    name = sa.Column(sa.Unicode(1024), nullable=False, unique=True)

    @might_commit
    @might_flush
    @classmethod
    def edit(cls, oid=None, **kwargs):
        obj = super(BundleMixin, cls).edit(oid=oid, _commit=False, **kwargs)
        return obj


class GroupMixin(object):
    """Generic mixin for user groups."""
    name = sa.Column(sa.Unicode(1024), nullable=False, unique=True)

    def get_all_permissions(self):
        perm_cls = registry().permission_cls
        mapping = self._query_permission_mapping().alias('group_permissions_mapping')
        q = db.session.query(
            perm_cls
        ).select_from(
            perm_cls
        ).join(
            mapping,
            mapping.c.perm_id == perm_cls.id
        ).filter(
            mapping.c.group_id == self.id
        )
        return set(q)

    @classmethod
    def _query_permission_mapping(cls):
        perm_cls = registry().permission_cls
        bundle_cls = registry().bundle_cls

        direct = db.session.query(
            cls.id.label('group_id'),
            perm_cls.id.label('perm_id')
        ).join(
            cls.permissions
        )

        via_bundle = db.session.query(
            cls.id.label('group_id'),
            perm_cls.id.label('perm_id')
        ).join(
            cls.bundles
        ).join(
            bundle_cls.permissions
        )
        return sa.union(direct, via_bundle)


class AttemptMixin(object):
    """Generic mixin for logging user login attempts."""
    # Form input data, e.g. username
    user_input = sa.Column(sa.Unicode(512), nullable=False)

    datetime_utc = sa.Column(ArrowType, nullable=False, default=arrow.utcnow,
                             server_default=dbutils.utcnow())
    attempt_type = sa.Column(AttemptType.db_type())
    is_during_lockout = sa.Column(sa.Boolean, nullable=False, default=False)
    success = sa.Column(sa.Boolean, nullable=False, default=True)
    source_ip = sa.Column(sa.Unicode(50), nullable=True)

    @classmethod
    def purge_attempts(cls, username=None, older_than=None, attempt_type=None):
        """Delete attempt records optionally filtered by username, age, or type."""
        query = cls.query
        if username:
            query = query.filter_by(user_input=username)

        if older_than:
            query = query.filter(cls.datetime_utc < arrow.utcnow().shift(days=-older_than))

        if attempt_type:
            query = query.filter_by(attempt_type=attempt_type)

        count = query.delete()
        db.session.commit()
        return count


def get_username(user):
    """Based on the registered user entity, find the column representing the login ID."""
    user_cls = registry().get_entity_cls('user')
    return getattr(user, get_username_key(user_cls))


def get_username_key(user_cls):
    obj = user_cls.username
    if not isinstance(obj, (sa.Column, sa_orm.attributes.InstrumentedAttribute)):
        obj = obj.descriptor.expr(user_cls)
    return obj.key


def _make_mapping_table(table_name, **foreign_cols):
    columns = (
        sa.Column(key, fc.type, sa.ForeignKey(fc, ondelete='CASCADE'), nullable=False,
                  primary_key=True)
        for key, fc in foreign_cols.items()
    )
    return db.Table(
        table_name,
        *columns
    )


def user_permission_mapping(user_cls, permission_cls, table_name='user_permissions',
                            user_id_attr='id', permission_id_attr='id',
                            user_rel_property='permissions'):
    table = _make_mapping_table(
        table_name,
        user_id=getattr(user_cls, user_id_attr),
        permission_id=getattr(permission_cls, permission_id_attr)
    )
    if user_rel_property:
        setattr(
            user_cls, user_rel_property,
            sa.orm.relationship(permission_cls, secondary=table)
        )

    return table


def bundle_permission_mapping(bundle_cls, permission_cls, table_name='bundle_permissions',
                              bundle_id_attr='id', permission_id_attr='id',
                              rel_property='permissions'):
    table = _make_mapping_table(
        table_name,
        bundle_id=getattr(bundle_cls, bundle_id_attr),
        permission_id=getattr(permission_cls, permission_id_attr)
    )

    if rel_property:
        setattr(
            bundle_cls, rel_property,
            sa.orm.relationship(permission_cls, secondary=table)
        )
    return table


def user_bundle_mapping(user_cls, bundle_cls, table_name='user_bundles',
                        user_id_attr='id', bundle_id_attr='id',
                        rel_property='bundles'):
    table = _make_mapping_table(
        table_name,
        user_id=getattr(user_cls, user_id_attr),
        bundle_id=getattr(bundle_cls, bundle_id_attr),
    )
    if rel_property:
        setattr(
            user_cls, rel_property,
            sa.orm.relationship(bundle_cls, secondary=table, backref='users')
        )
    return table


def user_group_mapping(user_cls, group_cls, table_name='user_groups',
                       user_id_attr='id', group_id_attr='id',
                       rel_property='groups'):
    table = _make_mapping_table(
        table_name,
        user_id=getattr(user_cls, user_id_attr),
        group_id=getattr(group_cls, group_id_attr),
    )
    if rel_property:
        setattr(
            user_cls, rel_property,
            sa.orm.relationship(group_cls, secondary=table, backref='users')
        )
    return table


def group_permission_mapping(group_cls, permission_cls, table_name='group_permissions',
                             group_id_attr='id', permission_id_attr='id',
                             rel_property='permissions'):
    table = _make_mapping_table(
        table_name,
        group_id=getattr(group_cls, group_id_attr),
        permission_id=getattr(permission_cls, permission_id_attr),
    )
    if rel_property:
        setattr(
            group_cls, rel_property,
            sa.orm.relationship(permission_cls, secondary=table)
        )
    return table


def group_bundle_mapping(group_cls, bundle_cls, table_name='group_bundles',
                         group_id_attr='id', bundle_id_attr='id',
                         rel_property='bundles'):
    table = _make_mapping_table(
        table_name,
        group_id=getattr(group_cls, group_id_attr),
        bundle_id=getattr(bundle_cls, bundle_id_attr),
    )
    if rel_property:
        setattr(
            group_cls, rel_property,
            sa.orm.relationship(bundle_cls, secondary=table, backref='groups')
        )
    return table


def initialize_mappings(namespace='keg_auth', registry=None):
    def _make_table_name(default_name):
        return '{}_{}'.format(namespace, default_name) if namespace else default_name

    mappings = {
        'user_permissions': (user_permission_mapping, 'user', 'permission'),
        'bundle_permissions': (bundle_permission_mapping, 'bundle', 'permission'),
        'user_bundles': (user_bundle_mapping, 'user', 'bundle'),
        'user_groups': (user_group_mapping, 'user', 'group'),
        'group_permissions': (group_permission_mapping, 'group', 'permission'),
        'group_bundles': (group_bundle_mapping, 'group', 'bundle')
    }
    tables = {}
    for base_name, mapping_data in mappings.items():
        table_func, type1, type2 = mapping_data
        table1 = registry.get_entity_cls(type1)
        table2 = registry.get_entity_cls(type2)

        tables[base_name] = table_func(table1, table2, table_name=_make_table_name(base_name))

    return tables


def initialize_events(registry=None):
    # look for changes to rights throughout users, groups, and bundles before flush. Reset the
    #   session key when there is a change
    def _isinstance(target, cls):
        # use a more simplistic method of determining type for performance
        return type(target) is cls

    def _sa_attr_has_changes(target, attr):
        try:
            return sa_orm.attributes.get_history(target, attr).has_changes()
        except KeyError as exc:
            if attr not in str(exc):
                raise
        return False

    @sa.event.listens_for(db.session, 'before_flush')
    def changed_users(session, *args):
        for target in session.new | session.dirty:
            if not _isinstance(target, registry.user_cls):
                continue

            if (
                _sa_attr_has_changes(target, 'permissions')
                or _sa_attr_has_changes(target, 'groups')
                or _sa_attr_has_changes(target, 'bundles')
                or _sa_attr_has_changes(target, 'is_superuser')
                or _sa_attr_has_changes(target, 'is_enabled')
            ):
                target.reset_session_key()

    @sa.event.listens_for(db.session, 'before_flush')
    def changed_groups(session, *args):
        for target in session.new | session.dirty:
            if not _isinstance(target, registry.group_cls):
                continue

            if (
                _sa_attr_has_changes(target, 'permissions')
                or _sa_attr_has_changes(target, 'bundles')
            ):
                for user in target.users:
                    user.reset_session_key()
            user_history = sa_orm.attributes.get_history(target, 'users')
            for user in user_history.added + user_history.deleted:
                user.reset_session_key()

        for target in session.deleted:
            if not _isinstance(target, registry.group_cls):
                continue

            for user in target.users:
                user.reset_session_key()

    @sa.event.listens_for(db.session, 'before_flush')
    def changed_bundles(session, *args):
        for target in session.new | session.dirty:
            if not _isinstance(target, registry.bundle_cls):
                continue

            if _sa_attr_has_changes(target, 'permissions'):
                for user in target.users:
                    user.reset_session_key()
                for group in target.groups:
                    for user in group.users:
                        user.reset_session_key()
            user_history = sa_orm.attributes.get_history(target, 'users')
            update_users = user_history.added + user_history.deleted

            group_history = sa_orm.attributes.get_history(target, 'groups')
            for group in group_history.added + group_history.deleted:
                update_users += tuple(group.users)

            for user in update_users:
                user.reset_session_key()

        for target in session.deleted:
            if not _isinstance(target, registry.bundle_cls):
                continue

            update_users = target.users
            for group in target.groups:
                update_users += group.users
            for user in update_users:
                user.reset_session_key()



import inspect

from keg_auth.extensions import lazy_gettext as _


class RegistryError(Exception):
    pass


class EntityRegistry(object):
    """Registry cache that identifies entities for particular usage/function in KegAuth.

    KegAuth does not provide its own entities for the database model. Instead, mixins are
    given so that an application can customize as needed. To support this model and still
    know what entity to use, we register it in an EntityRegistry.

    Entities may be registered in one of two ways:
    - Mark an entity with a registration decorator::

        @registry.register_user
        class User(UserMixin, db.EntityBase):
            pass

    - Pass the entity into the EntityRegistry constructor::

        registry = EntityRegistry(user=User, permission=Permission, ...)

    Registered entities may be subsequently referenced via the properties, e.g.

        registry.user_cls
    """
    def __init__(self, user=None, permission=None, bundle=None, group=None, attempt=None):
        self._user_cls = user
        self._permission_cls = permission
        self._bundle_cls = bundle
        self._group_cls = group
        self._attempt_cls = attempt

    def _type_to_attr(self, type):
        return '_{}_cls'.format(type)

    def register_entity(self, type, cls):
        attr = self._type_to_attr(type)
        try:
            if getattr(self, attr) is not None:
                raise RegistryError(_('Entity class already registered for {}').format(type))
        except AttributeError:
            raise RegistryError(_('Attempting to register unknown type {}').format(type))
        if not inspect.isclass(cls):
            raise RegistryError(_('Entity must be a class'))

        setattr(self, attr, cls)
        return cls

    def register_user(self, cls):
        """Mark given class as the entity for User."""
        return self.register_entity('user', cls)

    def register_permission(self, cls):
        """Mark given class as the entity for Permission."""
        return self.register_entity('permission', cls)

    def register_bundle(self, cls):
        """Mark given class as the entity for Bundle."""
        return self.register_entity('bundle', cls)

    def register_group(self, cls):
        """Mark given class as the entity for Group."""
        return self.register_entity('group', cls)

    def register_attempt(self, cls):
        """Mark given class as the entity for Attempt."""
        return self.register_entity('attempt', cls)

    def get_entity_cls(self, type):
        attr = self._type_to_attr(type)
        try:
            cls = getattr(self, attr)
        except AttributeError:
            raise RegistryError(_('Attempting to register unknown type {}').format(type))

        if cls is None:
            raise RegistryError(_('No entity registered for {}').format(type))
        return cls

    @property
    def user_cls(self):
        """Return the entity registered for User."""
        return self.get_entity_cls('user')

    @property
    def permission_cls(self):
        """Return the entity registered for Permission."""
        return self.get_entity_cls('permission')

    @property
    def bundle_cls(self):
        """Return the entity registered for Bundle."""
        return self.get_entity_cls('bundle')

    @property
    def group_cls(self):
        """Return the entity registered for Group."""
        return self.get_entity_cls('group')

    @property
    def attempt_cls(self):
        """Return the entity registered for Attempt."""
        return self.get_entity_cls('attempt')

    def is_registered(self, type):
        """Helper for determining if functionality is unlocked via a registered entity."""
        attr = self._type_to_attr(type)
        return getattr(self, attr, None) is not None


# Example
# @registry.register_user
# class User(db.EntityBase, UserMixin):
#     pass



import flask
import freezegun
import arrow
from keg_auth.core import update_last_login
from keg_auth_ta.model import entities as ents


class TestUpdateLastLogin(object):

    @freezegun.freeze_time("2018-10-01 15:00:00")
    def test_it_will_update_last_login(self):
        u = ents.User.testing_create(email=u"test@bar.com", last_login_utc=None)
        update_last_login(flask.current_app, u)
        ents.db.session.remove()
        u = ents.User.get_by(email=u"test@bar.com")
        assert u.last_login_utc == arrow.utcnow()



from __future__ import absolute_import

from flask_bootstrap import Bootstrap
from keg.app import Keg
from keg.db import db
from werkzeug.datastructures import ImmutableDict

from keg_auth_ta.extensions import auth_manager, csrf, mail_ext
from keg_auth_ta.grids import Grid
from keg_auth_ta.views import blueprints


class KegAuthTestApp(Keg):
    import_name = 'keg_auth_ta'
    db_enabled = True
    use_blueprints = blueprints
    keyring_enable = False
    visit_modules = ['.events']

    def on_init_complete(self):
        self.template_globals = ImmutableDict(
            auto_expand_menu=self.config.get('AUTO_EXPAND_MENU')
        )

        auth_manager.init_app(self)
        csrf.init_app(self)
        mail_ext.init_app(self)

        Grid.manager.init_db(db)
        Grid.manager.init_app(self)

        if Bootstrap is not None:
            Bootstrap(self)

        return self


if __name__ == '__main__':
    from keg_auth_ta import app
    app.KegAuthTestApp.cli.main()



import arrow
import mock
from blazeutils.containers import LazyDict
from keg.testing import CLIBase

from keg_auth.model.entity_registry import RegistryError
from keg_auth_ta.model import entities as ents


class TestCLI(CLIBase):

    def setup(self):
        ents.UserNoEmail.delete_cascaded()
        ents.User.delete_cascaded()
        ents.Attempt.delete_cascaded()

    def test_help_options(self):
        result = self.invoke('auth')
        assert 'create-user' in result.output

    @mock.patch('keg.current_app.auth_manager.create_user_cli', autospec=True, spec_set=True)
    def test_create_user_mock(self, m_cli_create_user):
        """ We mock in this test to more accurately test the URL creation. """
        user = LazyDict(id=3, _token_plain='1234')
        m_cli_create_user.return_value = user

        result = self.invoke('auth', 'create-user', 'foo@bar.com', 'abc', 'def')

        assert 'User created.\nEmail sent with verification URL.' in result.output
        assert 'Verification URL: http://keg.example.com/verify-account/3/1234' in result.output

        m_cli_create_user.assert_called_once_with(email='foo@bar.com', extra_args=('abc', 'def'),
                                                  is_superuser=False, mail_enabled=True)

    @mock.patch('keg.current_app.auth_manager.create_user_cli', autospec=True, spec_set=True)
    def test_create_superuser_mock(self, m_cli_create_user):
        """ We mock in this test to more accurately test the URL creation. """
        user = LazyDict(id=3, _token_plain='1234')
        m_cli_create_user.return_value = user

        result = self.invoke('auth', 'create-user', '--as-superuser', 'foo@bar.com')

        assert 'User created.\nEmail sent with verification URL.' in result.output
        assert 'Verification URL: http://keg.example.com/verify-account/3/1234' in result.output

        m_cli_create_user.assert_called_once_with(email='foo@bar.com', extra_args=(),
                                                  is_superuser=True, mail_enabled=True)

    @mock.patch('keg.current_app.auth_manager.mail_manager.send_new_user',
                autospec=True, spec_set=True)
    def test_create_user_integration(self, m_send):
        result = self.invoke('auth', 'create-user', 'foo@bar.com')

        output_parts = result.output.split('/')
        token = output_parts[-1].strip()
        user_id = output_parts[-2]

        assert 'User created.\nEmail sent with verification URL.' in result.output
        assert m_send.call_count

        user = ents.User.query.get(user_id)
        assert user.token_verify(token)

    @mock.patch('keg.current_app.auth_manager.mail_manager.send_new_user',
                autospec=True, spec_set=True)
    def test_create_user_user_args_integration(self, m_send):
        result = self.invoke('--profile', 'TestProfileUserArgs', 'auth', 'create-user',
                             'Bob Smith', 'bob@example.com')

        output_parts = result.output.split('/')
        user_id = output_parts[-2]

        user = ents.User.query.get(user_id)
        assert user.email == 'bob@example.com'
        assert user.name == 'Bob Smith'

        result = self.invoke('--profile', 'TestProfileUserArgs', 'auth', 'create-user',
                             'Joe Smith', 'joe@example.com')
        output_parts = result.output.split('/')
        user_id = output_parts[-2]

        user = ents.User.query.get(user_id)
        assert user.email == 'joe@example.com'
        assert user.name == 'Joe Smith'

    @mock.patch('keg.current_app.auth_manager.mail_manager', None)
    def test_create_user_no_mail_by_manager(self):
        result = self.invoke('auth', 'create-user', 'foo@bar.com')

        assert 'User created.' in result.output
        assert 'Email sent with verification URL.' not in result.output

    @mock.patch('keg.current_app.auth_manager.mail_manager.send_new_user',
                autospec=True, spec_set=True)
    def test_create_user_no_mail_by_flag(self, m_send):
        result = self.invoke('auth', 'create-user', '--no-mail', 'foo@bar.com')

        assert 'User created.' in result.output
        assert 'Email sent with verification URL.' not in result.output
        assert 'Verification URL: http://keg.example.com/verify-account' in result.output
        assert not m_send.call_count

    def test_set_password(self):
        ents.User.testing_create(email='test@level12.com')
        input_ = 'Hello123!\nHello123!'
        result = self.invoke('auth', 'set-password', 'test@level12.com', input=input_)
        assert result.stdout == 'Password: \nRepeat for confirmation: \n'

        assert ents.User.get_by(email='test@level12.com').password == 'Hello123!'

    def test_set_invalid_password(self):
        ents.User.testing_create(email='test@level12.com')
        input_ = 'Hello\nHello123!\nHello123!'
        result = self.invoke('auth', 'set-password', 'test@level12.com', input=input_)
        assert result.stdout == (
            'Password: \n'
            'Password does not meet the following restrictions:\n'
            '\t• Password must be at least 8 characters long\n'
            'Error: The value you entered was invalid.\n'
            'Password: \n'
            'Repeat for confirmation: \n'
        )

        assert ents.User.get_by(email='test@level12.com').password == 'Hello123!'

    @mock.patch('keg_auth.cli.click.echo', autospec=True, spec_set=True)
    def test_set_password_error(self, m_echo):
        self.invoke('auth', 'set-password', 'test@dne.com')
        m_echo.assert_called_once_with('Unknown user', err=True)

    def test_command_extension(self):
        result = self.invoke('auth', 'command-extension')

        assert 'verified' in result.output

    @mock.patch('keg.cli.click.echo', autospec=True, spec_set=True)
    def test_purge_attempts(self, m_echo):
        username = 'foo@bar.com'
        for i in range(0, 3):
            for attempt_type in ['login', 'reset']:
                for username in ['foo@test.com', 'bar@test.com']:
                    ents.Attempt.testing_create(
                        user_input=username,
                        attempt_type=attempt_type,
                        datetime_utc=arrow.utcnow().shift(days=-i)
                    )

        # Delete all attempts older than 2 days.
        assert ents.Attempt.query.count() == 12
        self.invoke('auth', 'purge-attempts', '--older-than=2')
        m_echo.assert_called_once_with('Deleted 4 attempts.')
        assert ents.Attempt.query.count() == 8

        # Delete all attempts for username.
        m_echo.reset_mock()
        self.invoke('auth', 'purge-attempts', '--username=foo@test.com')
        m_echo.assert_called_once_with('Deleted 4 attempts.')
        assert ents.Attempt.query.count() == 4

        # Delete all login attempts.
        m_echo.reset_mock()
        self.invoke('auth', 'purge-attempts', '--type=login')
        m_echo.assert_called_once_with('Deleted 2 attempts.')
        assert ents.Attempt.query.count() == 2

        # Delete all attempts.
        m_echo.reset_mock()
        self.invoke('auth', 'purge-attempts')
        m_echo.assert_called_once_with('Deleted 2 attempts.')
        assert ents.Attempt.query.count() == 0

    @mock.patch('keg.cli.click.echo', autospec=True, spec_set=True)
    @mock.patch('keg.current_app.auth_manager.entity_registry.get_entity_cls',
                autospec=True, spec_set=True, side_effect=RegistryError)
    def test_purge_attempts_no_attempt_registered(self, m_ent_registry, m_echo):
        self.invoke('auth', 'purge-attempts', '--username=foo@bar.com')
        m_echo.assert_called_once_with('No attempt class has been registered.')



import click
import keg

from keg_auth.model import get_username_key
from keg_auth.extensions import gettext as _
from keg_auth.libs.authenticators import PasswordPolicyError
from keg_auth.model.entity_registry import RegistryError


class PasswordType(click.ParamType):
    name = 'password'

    def __init__(self, policy, user):
        self.policy = policy
        self.user = user

    def convert(self, value, param, ctx):
        if not isinstance(value, str):
            self.fail(_('Password must be a string'), param, ctx)

        errors = []
        for check in self.policy.password_checks():
            try:
                check(value, self.user)
            except PasswordPolicyError as e:
                errors.append(str(e))

        if errors:
            error_list = '\n'.join('\t\N{BULLET} {}'.format(e) for e in errors)
            message = _('Password does not meet the following restrictions:\n{errs}',
                        errs=error_list)

            # because we hide the input, click also hides the error message, so echo manually
            click.echo(message)
            self.fail(
                message,
                param,
                ctx,
            )
        return value


def add_cli_to_app(app, cli_group_name, user_args=('email',)):

    @app.cli.group(cli_group_name)
    def auth():
        """ User and authentication related commands."""

    @auth.command('set-password', short_help='Set a user\'s password')
    @click.argument('username')
    def set_user_password(username):
        auth_manager = keg.current_app.auth_manager
        user_ent = auth_manager.entity_registry.user_cls
        user = user_ent.get_by(**{get_username_key(user_ent): username})

        if user is None:
            click.echo('Unknown user', err=True)
            return

        password_policy = auth_manager.password_policy_cls()
        password = click.prompt(
            'Password',
            type=PasswordType(password_policy, user),
            hide_input=True,
            confirmation_prompt=True
        )
        user.change_password(user.token_generate(), password)

    # note: no group attached here. We will apply the arguments and group it below
    @click.argument('extra_args', nargs=-1)
    @click.option('--as-superuser', is_flag=True)
    @click.option('--no-mail', is_flag=True)
    def _create_user(as_superuser, no_mail, **kwargs):
        """ Create a user.

            Create a user record with the given required args and (if a mail manager is
            configured) send them an email with URL and token to set their password.  Any
            EXTRA_ARGS will be sent to the auth manager for processing.
        """
        auth_manager = keg.current_app.auth_manager
        user = auth_manager.create_user_cli(is_superuser=as_superuser, mail_enabled=not no_mail,
                                            **kwargs)
        click.echo(_('User created.'))
        if auth_manager.mail_manager:
            if not no_mail:
                click.echo(_('Email sent with verification URL.'))
            verification_url = auth_manager.mail_manager.verify_account_url(user)
            click.echo(_('Verification URL: {url}').format(url=verification_url))

    # dress up _create_user as needed
    create_user = _create_user
    for arg in reversed(user_args):
        create_user = click.argument(arg)(create_user)
    auth.command('create-user')(create_user)

    @click.option('--username', '--user', help='username to filter by')
    @click.option('--older-than', type=int, help='number of days')
    @click.option('--attempt-type', '--type', help='[login, reset, forgot]')
    def purge_attempts(username, older_than, attempt_type):
        """Purge authentication attempts optionally filtered by username, type, or age."""
        auth_manager = keg.current_app.auth_manager
        try:
            attempt_ent = auth_manager.entity_registry.attempt_cls
        except RegistryError:
            click.echo('No attempt class has been registered.')
            return

        count = attempt_ent.purge_attempts(
            username=username,
            older_than=older_than,
            attempt_type=attempt_type
        )
        click.echo(f'Deleted {count} attempts.')

    auth.command('purge-attempts')(purge_attempts)

    app.auth_manager.cli_group = auth



# Using unicode_literals instead of adding 'u' prefix to all stings that go to SA.
from __future__ import unicode_literals

import flask
from keg_auth import mail
import mock

from keg_auth_ta.app import mail_ext
from keg_auth_ta.model import entities as ents


class TestMailTemplate(object):

    def test_normal(self):
        email = mail.mail_template('email.j2', name='foo')
        assert email.subject == '[KA Demo] Example Subject'
        assert email.text == '**Hello foo**'
        assert email.html == '<p><strong>Hello foo</strong></p>\n'

    def test_no_abbr(self):
        with mock.patch.dict(flask.current_app.config, KEGAUTH_EMAIL_SITE_ABBR=None):
            email = mail.mail_template('email.j2')
        assert email.subject == 'Example Subject'


class TestAuthMailManager(object):
    @classmethod
    def setup_class(cls):
        cls.mb = mail.AuthMailManager(mail_ext)

    def setup(self):
        ents.User.delete_cascaded()

    def test_reset_password_message(self):
        user = ents.User.testing_create(email='foo@bar.com')
        user.token_generate()
        link_url = flask.current_app.auth_manager.mail_manager.reset_password_url(user)

        with flask.current_app.test_request_context():
            message = self.mb.reset_password_message(user)

        assert message.subject == '[KA Demo] Password Reset Link'
        assert message.recipients == ['foo@bar.com']
        assert 'Somebody asked to reset your password on Keg Auth Demo.' in message.body
        assert '<{}>'.format(link_url) in message.body
        assert 'href="{}"'.format(link_url) in message.html

    def test_send_reset_password(self):
        user = ents.User.testing_create()

        with mail_ext.record_messages() as outbox:
            flask.current_app.auth_manager.mail_manager.send_reset_password(user)

            assert len(outbox) == 1
            assert outbox[0].subject == '[KA Demo] Password Reset Link'

    def test_new_user_message(self):
        user = ents.User.testing_create(email='foo@bar.com')
        user.token_generate()
        link_url = flask.current_app.auth_manager.mail_manager.verify_account_url(user)

        with flask.current_app.test_request_context():
            message = self.mb.new_user_message(user)

        assert message.subject == '[KA Demo] User Welcome & Verification'
        assert message.recipients == ['foo@bar.com']
        assert 'A user account has been created for you on Keg Auth Demo.' in message.body
        assert '<{}>'.format(link_url) in message.body
        assert 'href="{}"'.format(link_url) in message.html

    def test_send_new_user(self):
        user = ents.User.testing_create()

        with mail_ext.record_messages() as outbox:
            flask.current_app.auth_manager.mail_manager.send_new_user(user)

            assert len(outbox) == 1
            assert outbox[0].subject == '[KA Demo] User Welcome & Verification'



import inspect
import sys

from blazeutils.strings import simplify_string
import flask
import flask_login

from keg_auth.extensions import lazy_gettext as _
from keg_auth.model.utils import has_permissions

try:
    from speaklater import is_lazy_string
except ImportError:
    is_lazy_string = lambda value: False  # noqa: E731


def get_defining_class(func):
    if inspect.isclass(func):
        return

    if sys.version_info[0] == 2:
        return getattr(func, 'im_class', None)  # pragma: no cover

    if inspect.isfunction(func):
        parse_def = func.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)
        if len(parse_def) == 1:
            # looks like a method without a class
            return
        return getattr(inspect.getmodule(func), parse_def[0])


class NavURL(object):
    """Wraps url_for with permission-checking to determine if user should see a route.

    Endpoint is checked for user/permission requirements.
    - method/class/blueprint permissions from decorators (preferred in most cases)
    - `requires_permissions` kwarg specifies conditions and disregards the decorators
    - `requires_anonymous` kwarg forces nav for only unauthenticated users

    Note that permission requirements are checked at all levels of the view hierarchy as
    needed: method, class, and blueprint.
    """
    def __init__(self, route_string, *args, **kwargs):
        self.route_string = route_string
        self.route_args = args
        self.route_kwargs = kwargs
        self.requires_permissions = kwargs.pop('requires_permissions', None)
        self.requires_anonymous = kwargs.pop('requires_anonymous', None)

    @property
    def url(self):
        return flask.url_for(self.route_string, *self.route_args, **self.route_kwargs)

    @property
    def is_permitted(self):
        """ Check permitted status of this route for the current user """
        # simplest case: route has requirements directly assigned
        if self.requires_permissions:
            if not flask_login.current_user or not flask_login.current_user.is_authenticated:
                return False
            return has_permissions(
                self.requires_permissions,
                flask_login.current_user
            )

        # other simple case: route is forced to show only anonymous users
        if self.requires_anonymous:
            return not bool(
                flask_login.current_user
                and flask_login.current_user.is_authenticated
            )

        # otherwise, we need to find the view for the route. In that case, both the route and its
        #   defining class (if any) may (or may not) have requirements to check.
        # the following checks are ANDed, so return False if anything fails
        view_obj = flask.current_app.view_functions.get(self.route_string)
        if not view_obj:
            raise Exception(
                _('Endpoint {} in navigation is not registered').format(self.route_string)
            )

        def check_auth(obj):
            if obj is None:
                return True

            if (
                getattr(obj, '__keg_auth_requires_user__', False) and (
                    not flask_login.current_user
                    or not flask_login.current_user.is_authenticated
                )
            ):
                return False

            if (
                getattr(obj, '__keg_auth_requires_permissions__', False)
                and not has_permissions(
                    obj.__keg_auth_requires_permissions__,
                    flask_login.current_user
                )
            ):
                return False

            return True

        def fetch_parent_class(view_obj):
            parent_class = getattr(
                view_obj, 'im_class',
                getattr(view_obj, '__keg_auth_parent_class__', None)
            )
            if not parent_class and not hasattr(view_obj, '__keg_auth_parent_class__'):
                obj = view_obj

                if hasattr(obj, '__keg_auth_original_function__'):
                    # the target method has been wrapped by a keg auth decorator, so we need
                    #   to inspect the original method to find the parent class (if any)
                    obj = obj.__keg_auth_original_function__

                view_obj.__keg_auth_parent_class__ = get_defining_class(obj)
                parent_class = view_obj.__keg_auth_parent_class__
            return parent_class

        def fetch_blueprint():
            return flask.current_app.blueprints.get(self.route_string.split('.', 1)[0], None)

        if hasattr(view_obj, 'view_class'):
            # class got wrapped with flask's as_view - get the original view to see what
            #   requirements are stored there
            view_obj = view_obj.view_class

        if inspect.isclass(view_obj) and hasattr(view_obj, 'get'):
            # view class has an action method likely to be called via a navigation link
            if sys.version_info[0] != 2:
                view_obj.get.__keg_auth_parent_class__ = view_obj
            view_obj = view_obj.get

        # make sure defining class is assigned (if any). We need to know this in order to
        #   check requirements at the class level
        parent_class = fetch_parent_class(view_obj)

        blueprint = fetch_blueprint()

        return check_auth(view_obj) and check_auth(parent_class) and check_auth(blueprint)


class NavItem(object):
    """Defines a menu item or structure of a menu.

    Example::

        my_menu = NavItem(
            NavItem(
                'Admin',
                NavItem('Users', NavURL('auth.user:list')),
                NavItem('Groups', NavURL('auth.group:list')),
                nav_group='admin',
                icon_class='fas fa-briefcase'
            ),
            NavItem(
                'Reports',
                NavItem('Frequency', NavURL('frequency-report')),
                NavItem('Financial', NavURL('money-report', requires_permissions='secret-perm'))
            )
        )

    """
    class NavItemType(object):
        STEM = 0
        LEAF = 1

    def __init__(self, *args, nav_group=None, icon_class=None):
        self.label = None
        if len(args) and (isinstance(args[0], str) or is_lazy_string(args[0])):
            self.label = args[0]
            args = args[1:]
        self.route = None
        self.sub_nodes = None
        self.nav_group = nav_group
        self.icon_class = icon_class

        # cache permission-related items
        self._is_permitted = {}
        self._permitted_sub_nodes = {}

        if len(args) == 0:
            raise Exception(_('must provide a NavURL or a list of NavItems'))

        if isinstance(args[0], NavURL):
            self.route = args[0]
            if len(args) > 1:
                args = args[1:]
            else:
                return

        if len(args):
            self.sub_nodes = args
            if not self.nav_group:
                self.nav_group = simplify_string(self.label or '__root__')

    def clear_authorization(self, session_key):
        """Reset cached authorization in this and all subnodes for the given session key."""
        self._is_permitted.pop(session_key, None)
        self._permitted_sub_nodes.pop(session_key, None)
        for sub_node in (self.sub_nodes or []):
            sub_node.clear_authorization(session_key)

    @property
    def node_type(self):
        """Return type NavItemType indicating whether this node is at the end of the structure."""
        if self.sub_nodes:
            return NavItem.NavItemType.STEM
        return NavItem.NavItemType.LEAF

    @property
    def is_permitted(self):
        """Compute/cache authorization from permission conditions, and return bool."""
        current_user = flask_login.current_user
        session_key = current_user.get_id() if current_user else None
        if self._is_permitted.get(session_key) is None:
            if self.node_type == NavItem.NavItemType.LEAF:
                # checks the route for requirements, or the target view/class
                self._is_permitted[session_key] = self.route.is_permitted
            else:
                # find a subnode that is permitted
                self._is_permitted[session_key] = (len(self.permitted_sub_nodes) > 0)

        return self._is_permitted.get(session_key)

    @property
    def permitted_sub_nodes(self):
        """Return list of subnodes accessible to current user."""
        current_user = flask_login.current_user
        session_key = current_user.get_id() if current_user else None
        if self._permitted_sub_nodes.get(session_key) is None:
            self._permitted_sub_nodes[session_key] = [
                node for node in (self.sub_nodes or []) if node.is_permitted
            ]

        return self._permitted_sub_nodes.get(session_key)

    @property
    def has_current_route(self):
        """Returns true if current request matches this nav node."""
        if self.route:
            return self.route.route_string == flask.request.endpoint
        else:
            for node in self.permitted_sub_nodes:
                if node.has_current_route:
                    return True
        return False



from __future__ import unicode_literals

import flask
import mock
from keg_auth.core import AuthManager
from keg_auth.libs.authenticators import KegAuthenticator, JwtRequestLoader

from keg_auth_ta.app import mail_ext
from keg_auth_ta.extensions import auth_entity_registry
from keg_auth_ta.model import entities as ents


class TestAuthManager(object):
    def setup(self):
        ents.Permission.delete_cascaded()
        ents.User.delete_cascaded()
        self.am = flask.current_app.auth_manager

    def test_create_user(self):
        with mail_ext.record_messages() as outbox:
            user = self.am.create_user(dict(email=u'foo@bar.com'))

        assert len(outbox) == 1
        assert outbox[0].subject == '[KA Demo] User Welcome & Verification'

        assert user.email == 'foo@bar.com'
        assert user._token_plain
        assert ents.User.query.count() == 1

    def test_create_user_no_commit(self):
        self.am.create_user(dict(email=u'foo@bar.com'), _commit=False)
        ents.db.session.rollback()
        assert ents.User.query.count() == 0

    def test_create_user_commit(self):
        self.am.create_user(dict(email=u'foo@bar.com'), _commit=True)
        ents.db.session.rollback()
        assert ents.User.query.count() == 1

    @mock.patch('keg_auth.core.model.initialize_mappings')
    def test_model_initialized_only_once(self, m_init):
        self.am.init_app(flask.current_app)
        assert not m_init.called

    def test_permissions_synced_to_db(self):
        # create a permission that will get destroyed by sync, and ensure no integrity errors
        permission_to_delete = ents.Permission.add(token='snoopy')
        ents.Group.testing_create(permissions=[permission_to_delete])
        ents.Bundle.testing_create(permissions=[permission_to_delete])
        ents.User.testing_create(permissions=[permission_to_delete])

        # token should not be duplicated during sync, but description should be set
        ents.Permission.add(token='bar')

        # define the app permissions
        permissions = (
            'foo',
            ('bar', 'location for libations'),
            ('baz', 'nonsense word'),
        )
        with mock.patch.object(self.am, 'permissions', permissions):
            # we want to make sure the sync will still
            # run when mode is not testing
            flask.current_app.config['TESTING'] = False
            self.am.init_app(flask.current_app)

        assert ents.Permission.get_by(token='foo')
        assert ents.Permission.get_by(token='bar').description == 'location for libations'
        assert ents.Permission.get_by(token='baz').description == 'nonsense word'
        assert not ents.Permission.get_by(token='snoopy')
        flask.current_app.config['TESTING'] = True

    @mock.patch('keg_auth.core.KegAuthenticator')
    def test_request_loaders_initialized_only_once(self, m_init):
        self.am.init_app(flask.current_app)
        assert not m_init.called

    @mock.patch('keg_auth.core.AuthManager.init_model')
    def test_request_loaders_initialized(self, m_model):
        app = mock.MagicMock()
        manager = AuthManager(None, request_loaders=[JwtRequestLoader],
                              entity_registry=auth_entity_registry)
        manager.init_app(app)
        assert isinstance(manager.login_authenticator, KegAuthenticator)
        assert isinstance(manager.get_request_loader('jwt'), JwtRequestLoader)

    def test_resend_verification(self):
        user = ents.User.testing_create(
            email='foo1@bar.com'
        )
        with mail_ext.record_messages() as outbox:
            self.am.resend_verification_email(user.id)

        assert user._token_plain

        assert len(outbox) == 1
        assert outbox[0].subject == '[KA Demo] User Welcome & Verification'
        assert '/verify-account/{}/{}'.format(user.id, user._token_plain) in outbox[0].body
